from typing import Optional, TYPE_CHECKING, Union

from ..util import prettify, get_optional, get_required, fromisoformat
from ..enums import (
    NotificationCategory,
    ObjectType,
    NotificationType,
    ChatChannelType,
    GameModeStr,
)


if TYPE_CHECKING:
    from datetime import datetime


__all__ = (
    "Notification",
    "ReadNotification",
    "NotificationsDetailsBase",
    "BeatmapOwnerChangeDetails",
    "BeatmapsetNotificationDetails",
    "BeatmapsetDiscussionPostNotificationDetails",
    "ReviewStats",
    "BeatmapsetDiscussionReviewNewDetails",
    "ChannelAnnouncementDetails",
    "ChannelMessageDetails",
    "CommentNewDetails",
    "ForumTopicReplyDetails",
    "UserAchievementUnlockDetails",
    "UserBeatmapsetNewDetails",
    "BeatmapsetDiscussionLockDetails",
    "BeatmapsetDiscussionPostNewDetails",
    "BeatmapsetDiscussionQualifiedProblemDetails",
    "BeatmapsetDiscussionUnlockDetails",
    "BeatmapsetDisqualifyDetails",
    "BeatmapsetLoveDetails",
    "BeatmapsetNominateDetails",
    "BeatmapsetQualifyDetails",
    "BeatmapsetRankDetails",
    "BeatmapsetRemoveFromLovedDetails",
    "BeatmapsetResetNominationsDetails",
    "UserBeatmapsetReviveDetails",
)


class Notification:
    """
    Represents a notification object.
    Go to :class:`Details` object to see details for each event

    **Attributes**

    id: int

    name: :class:`NotificationType`
        The type of event

    created_at: :py:class:`datetime.datetime`

    object_type: :class:`ObjectType`

    object_id: int

    source_user_id: Optional[int]

    is_read: bool

    details: Union[
    :class:`BeatmapOwnerChangeDetails`, :class:`BeatmapsetDiscussionLockDetails`,
    :class:`BeatmapsetDiscussionPostNewDetails`, :class:`BeatmapsetDiscussionQualifiedProblemDetails`,
    :class:`BeatmapsetDiscussionReviewNewDetails`, :class:`BeatmapsetDiscussionUnlockDetails`,
    :class:`BeatmapsetDisqualifyDetails`, :class:`BeatmapsetLoveDetails`,
    :class:`BeatmapsetNominateDetails`, :class:`BeatmapsetQualifyDetails`,
    :class:`BeatmapsetRankDetails`, :class:`BeatmapsetRemoveFromLovedDetails`,
    :class:`BeatmapsetResetNominationsDetails`, :class:`ChannelAnnouncementDetails`,
    :class:`ChannelMessageDetails`, :class:`CommentNewDetails`,
    :class:`ForumTopicReplyDetails`, :class:`UserAchievementUnlockDetails`,
    :class:`UserBeatmapsetNewDetails`, :class:`UserBeatmapsetReviveDetails`
    ]

        Details of the notification.

    **Documented event names and their attribute meanings**

    beatmapset_discussion_lock
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who locked discussion

    beatmapset_discussion_post_new
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: Poster of the discussion

    beatmapset_discussion_unlock
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who unlocked discussion

    beatmapset_disqualify
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who disqualified beatmapset

    beatmapset_love
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who promoted beatmapset to Loved

    beatmapset_nominate
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who nominated beatmapset

    beatmapset_qualify
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User whom beatmapset nomination triggered qualification

    beatmapset_remove_from_loved
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who removed beatmapset from Loved

    beatmapset_reset_nominations
        object_id: Beatmapset id
        object_type: beatmapset
        source_user_id: User who locked discussion

    channel_message
        object_id: Channel id
        object_type: channel
        source_user_id: User who posted message

    forum_topic_reply
        object_id: Topic id
        object_type: forum_topic
        source_user_id: User who posted message
    """

    __slots__ = (
        "id",
        "name",
        "created_at",
        "object_type",
        "object_id",
        "is_read",
        "source_user_id",
        "details",
    )

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.name: NotificationType = NotificationType(get_required(data, "name"))
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.object_type: ObjectType = ObjectType(get_required(data, "object_type"))
        self.object_id: int = get_required(data, "object_id")
        self.source_user_id: Optional[int] = data.get("source_user_id")
        self.is_read: bool = get_required(data, "is_read")
        self.details: _DETAILS_TYPE = _get_details_object(get_required(data, "details"), self.name)

    def __repr__(self):
        return prettify(self, "name", "details")


class ReadNotification:
    """
    Represents a read notification.

    **Attributes**

    category: :class:`NotificationCategory`

    id: int

    object_id: int

    object_type: :class:`ObjectType`
    """

    __slots__ = ("category", "id", "object_id", "object_type")

    def __init__(self, data):
        self.category: NotificationCategory = NotificationCategory(get_required(data, "category"))
        self.id: int = get_required(data, "id")
        self.object_id: int = get_required(data, "object_id")
        self.object_type: ObjectType = ObjectType(get_required(data, "object_type"))

    def __repr__(self):
        return prettify(self, "id")


class NotificationsDetailsBase:
    """
    Base class for all Detail objects.

    **Attributes**

    username: str
    """

    __slots__ = ("username",)

    def __init__(self, data):
        self.username: str = get_required(data, "username")

    def __repr__(self):
        # all the subclasses have a title attribute
        return prettify(self, "title", "username")


class BeatmapOwnerChangeDetails(NotificationsDetailsBase):
    """
    When ownership of a beatmap is transferred (for guest difficulties).

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    beatmap_id: int

    cover_url: str

    title: str

    title_unicode: str

    version: str
    """

    __slots__ = ("beatmap_id", "cover_url", "title", "title_unicode", "version")

    def __init__(self, data):
        super().__init__(data)
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.cover_url: str = get_required(data, "cover_url")
        self.title: str = get_required(data, "title")
        self.title_unicode: str = get_required(data, "title_unicode")
        self.version: str = get_required(data, "version")


class BeatmapsetNotificationDetails(NotificationsDetailsBase):
    """
    Base class for several other detail objects.

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: str

    title_unicode: str

    cover_url: str
    """

    __slots__ = ("title", "title_unicode", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.title: str = get_required(data, "title")
        self.title_unicode: str = get_required(data, "title_unicode")
        self.cover_url: str = get_required(data, "cover_url")


class BeatmapsetDiscussionPostNotificationDetails(NotificationsDetailsBase):
    """
    Base class for a couple other detail objects.

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    content: str

    title: str

    title_unicode: str

    post_id: int

    discussion_id: int

    beatmap_id: int

    cover_url: str
    """

    __slots__ = (
        "content",
        "title",
        "title_unicode",
        "post_id",
        "discussion_id",
        "beatmap_id",
        "cover_url",
    )

    def __init__(self, data):
        super().__init__(data)
        self.content: str = get_required(data, "content")
        self.title: str = get_required(data, "title")
        self.title_unicode: str = get_required(data, "title_unicode")
        self.post_id: int = get_required(data, "post_id")
        self.discussion_id: int = get_required(data, "discussion_id")
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.cover_url: str = get_required(data, "cover_url")


class ReviewStats:
    """
    **Attributes**

    praises: int

    suggestions: int

    problems: int
    """

    __slots__ = ("praises", "suggestions", "problems")

    def __init__(self, data):
        self.praises = get_required(data, "praises")
        self.suggestions = get_required(data, "suggestions")
        self.problems = get_required(data, "problems")


class BeatmapsetDiscussionReviewNewDetails(NotificationsDetailsBase):
    """
    New beatmapset discussion review.

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: str

    title_unicode: str

    post_id: int

    discussion_id: int

    beatmap_id: int

    cover_url: str

    embeds: :class:`ReviewStats`
    """

    __slots__ = (
        "title",
        "title_unicode",
        "post_id",
        "discussion_id",
        "beatmap_id",
        "cover_url",
        "embeds",
    )

    def __init__(self, data):
        super().__init__(data)
        self.title: str = get_required(data, "title")
        self.title_unicode: str = get_required(data, "title_unicode")
        self.post_id: int = get_required(data, "post_id")
        self.discussion_id: int = get_required(data, "discussion_id")
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.cover_url: str = get_required(data, "cover_url")
        self.embeds: ReviewStats = ReviewStats(get_required(data, "embeds"))


class ChannelAnnouncementDetails(NotificationsDetailsBase):
    """
    Chat channel announcement

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    channel_id: int

    name: str

    title: str

    type: :class:`ChatChannelType`

    cover_url: str
    """

    __slots__ = ("channel_id", "name", "title", "type", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.channel_id: int = get_required(data, "channel_id")
        self.name: str = get_required(data, "name")
        self.title: str = get_required(data, "title")
        self.type: ChatChannelType = ChatChannelType(get_required(data, "type").upper())
        self.cover_url: str = get_required(data, "cover_url")


class ChannelMessageDetails(NotificationsDetailsBase):
    """
    Chat channel message

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: str

    type: :class:`ChatChannelType`

    cover_url: str
    """

    __slots__ = ("title", "type", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.title: str = get_required(data, "title")
        self.type: ChatChannelType = ChatChannelType(get_required(data, "type").upper())
        self.cover_url: str = get_required(data, "cover_url")


class CommentNewDetails(NotificationsDetailsBase):
    """
    New comment

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    comment_id: int

    title: str

    content: str

    cover_url: str
    """

    __slots__ = ("comment_id", "title", "content", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.comment_id: int = get_required(data, "comment_id")
        self.title: str = get_required(data, "title")
        self.content: str = get_required(data, "content")
        self.cover_url: str = get_required(data, "cover_url")


class ForumTopicReplyDetails(NotificationsDetailsBase):
    """
    Forum topic reply

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    title: str

    post_id: int

    cover_url: str
    """

    __slots__ = ("title", "post_id", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.title: str = get_required(data, "title")
        self.post_id: int = get_required(data, "post_id")
        self.cover_url: str = get_required(data, "cover_url")


class UserAchievementUnlockDetails(NotificationsDetailsBase):
    """
    New achievement unlocked

    Extends :class:`NotificationsDetailsBase`

    **Attributes**

    achievement_id: int

    achievement_mode: Optional[:class:`GameModeStr`]

    cover_url: str

    slug: str

    title: str

    user_id: int
    """

    __slots__ = (
        "achievement_id",
        "achievement_mode",
        "cover_url",
        "slug",
        "title",
        "user_id",
    )

    def __init__(self, data):
        super().__init__(data)
        self.achievement_id: int = get_required(data, "achievement_id")
        self.achievement_mode: Optional[GameModeStr] = get_optional(data, "achievement_mode", GameModeStr)
        self.cover_url: str = get_required(data, "cover_url")
        self.slug: str = get_required(data, "slug")
        self.title: str = get_required(data, "title")
        self.user_id: int = get_required(data, "user_id")


class UserBeatmapsetNewDetails(NotificationsDetailsBase):
    """
    New beatmapset

    Extends: :class:`NotificationsDetailsBase`

    **Attributes**

    beatmapset_id: int

    title: str

    title_unicode: str

    cover_url: str
    """

    __slots__ = ("beatmapset_id", "title", "title_unicode", "cover_url")

    def __init__(self, data):
        super().__init__(data)
        self.beatmapset_id: int = get_required(data, "beatmapset_id")
        self.title: str = get_required(data, "title")
        self.title_unicode: str = get_required(data, "title_unicode")
        self.cover_url: str = get_required(data, "cover_url")


class BeatmapsetDiscussionLockDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset discussion locked

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetDiscussionPostNewDetails(BeatmapsetDiscussionPostNotificationDetails):
    """
    New beatmapset discussion post

    Extends :class:`BeatmapsetDiscussionPostNotificationDetails`
    """


class BeatmapsetDiscussionQualifiedProblemDetails(BeatmapsetDiscussionPostNotificationDetails):
    """
    Beatmapset discussion qualified problem

    Extends :class:`BeatmapsetDiscussionPostNotification`
    """


class BeatmapsetDiscussionUnlockDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset discussion unlocked

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetDisqualifyDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset disqualified

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetLoveDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset loved

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetNominateDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset nominated

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetQualifyDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset qualified

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetRankDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset ranked

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetRemoveFromLovedDetails(BeatmapsetNotificationDetails):
    """
    Beatmapset removed from love

    Extends :class:`BeatmapsetNotificationDetails`
    """


class BeatmapsetResetNominationsDetails(BeatmapsetNominateDetails):
    """
    Beatmapset nominations reset

    Extends :class:`BeatmapsetNominateDetails`
    """


class UserBeatmapsetReviveDetails(UserBeatmapsetNewDetails):
    """
    User beatmapset revived

    Extends :class:`UserBeatmapsetNewDetails`
    """


_DETAILS_TYPE = Union[
    BeatmapOwnerChangeDetails,
    BeatmapsetDiscussionLockDetails,
    BeatmapsetDiscussionPostNewDetails,
    BeatmapsetDiscussionQualifiedProblemDetails,
    BeatmapsetDiscussionReviewNewDetails,
    BeatmapsetDiscussionUnlockDetails,
    BeatmapsetDisqualifyDetails,
    BeatmapsetLoveDetails,
    BeatmapsetNominateDetails,
    BeatmapsetQualifyDetails,
    BeatmapsetRankDetails,
    BeatmapsetRemoveFromLovedDetails,
    BeatmapsetResetNominationsDetails,
    ChannelAnnouncementDetails,
    ChannelMessageDetails,
    CommentNewDetails,
    ForumTopicReplyDetails,
    UserAchievementUnlockDetails,
    UserBeatmapsetNewDetails,
    UserBeatmapsetReviveDetails,
]
_DETAIL_MAP = {
    NotificationType.BEATMAP_OWNER_CHANGE: BeatmapOwnerChangeDetails,
    NotificationType.BEATMAPSET_DISCUSSION_LOCK: BeatmapsetDiscussionLockDetails,
    NotificationType.BEATMAPSET_DISCUSSION_POST_NEW: BeatmapsetDiscussionPostNewDetails,
    NotificationType.BEATMAPSET_DISCUSSION_QUALIFIED_PROBLEM: BeatmapsetDiscussionQualifiedProblemDetails,
    NotificationType.BEATMAPSET_DISCUSSION_REVIEW_NEW: BeatmapsetDiscussionReviewNewDetails,
    NotificationType.BEATMAPSET_DISCUSSION_UNLOCK: BeatmapsetDiscussionUnlockDetails,
    NotificationType.BEATMAPSET_DISQUALIFY: BeatmapsetDisqualifyDetails,
    NotificationType.BEATMAPSET_LOVE: BeatmapsetLoveDetails,
    NotificationType.BEATMAPSET_NOMINATE: BeatmapsetNominateDetails,
    NotificationType.BEATMAPSET_QUALIFY: BeatmapsetQualifyDetails,
    NotificationType.BEATMAPSET_RANK: BeatmapsetRankDetails,
    NotificationType.BEATMAPSET_REMOVE_FROM_LOVED: BeatmapsetRemoveFromLovedDetails,
    NotificationType.BEATMAPSET_RESET_NOMINATIONS: BeatmapsetResetNominationsDetails,
    NotificationType.CHANNEL_ANNOUNCEMENT: ChannelAnnouncementDetails,
    NotificationType.CHANNEL_MESSAGE: ChannelMessageDetails,
    NotificationType.COMMENT_NEW: CommentNewDetails,
    NotificationType.FORUM_TOPIC_REPLY: ForumTopicReplyDetails,
    NotificationType.USER_ACHIEVEMENT_UNLOCK: UserAchievementUnlockDetails,
    NotificationType.USER_BEATMAPSET_NEW: UserBeatmapsetNewDetails,
    NotificationType.USER_BEATMAPSET_REVIVE: UserBeatmapsetReviveDetails,
}


def _get_details_object(data, detail_type) -> _DETAILS_TYPE:
    return _DETAIL_MAP[detail_type](data)



from typing import Optional, TYPE_CHECKING

from ..util import prettify, get_optional, get_required, fromisoformat

if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("NewsPost", "Navigation")


class NewsPost:
    """
    **Attributes**

    author: str

    edit_url: str
        Link to the file view on GitHub.

    first_image: Optional[str]
        Link to the first image in the document.

    first_image_2x: Optional[str]
        2x version of `first_image`. Will be `None` if `first_image` is `None` and vice-versa.

    id: int

    published_at: :py:class:`datetime.datetime`

    slug: str
        Filename without the extension, used in URLs.

    title: str

    updated_at: :py:class:`datetime.datetime`

    content: Optional[str]
        HTML post content.

    navigation: Optional[:class:`Navigation`]
        Navigation metadata.

    preview: Optional[str]
        First paragraph of content with HTML markup stripped.
    """

    __slots__ = (
        "author",
        "edit_url",
        "first_image",
        "first_image_2x",
        "id",
        "published_at",
        "slug",
        "title",
        "updated_at",
        "content",
        "navigation",
        "preview",
    )

    def __init__(self, data):
        self.author: str = get_required(data, "author")
        self.edit_url: str = get_required(data, "edit_url")
        self.first_image: Optional[str] = get_required(data, "first_image")
        self.first_image_2x: Optional[str] = get_required(data, "first_image@2x")
        self.id: int = get_required(data, "id")
        self.published_at: datetime = fromisoformat(get_required(data, "published_at"))
        self.slug: str = get_required(data, "slug")
        self.title: str = get_required(data, "title")
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))
        self.content: Optional[str] = data.get("content")
        self.navigation: Optional[Navigation] = get_optional(data, "navigation", Navigation)
        self.preview: Optional[str] = data.get("preview")

    def __repr__(self):
        return prettify(self, "title")


class Navigation:
    """
    **Attributes**

    newer: Optional[:class:`NewsPost`]
        null if the next post is not present.

    older: Optional[:class:`NewsPost`]
        null if the previous post is not present.
    """

    __slots__ = ("newer", "older")

    def __init__(self, data):
        self.newer: Optional[NewsPost] = get_optional(data, "newer", NewsPost)
        self.older: Optional[NewsPost] = get_optional(data, "older", NewsPost)

    def __repr__(self):
        return prettify(self, "newer", "older")



from typing import Optional, List, TYPE_CHECKING

from ..util import prettify, get_optional, get_required, fromisoformat
from ..enums import CommentSort
from .user import UserCompact
from .current_user_attributes import CommentableMetaAttributes


if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("Comment", "CommentBundle", "CommentableMeta")


class Comment:
    """
    Represents a single comment.

    **Attributes**

    commentable_id: int
        ID of the object the comment is attached to

    commentable_type: str
        type of object the comment is attached to

    created_at: :py:class:`datetime.datetime`

    deleted_at: Option[:py:class:`datetime.datetime`]

    deleted_by_id: Optional[int]
        user id of the user that deleted the post; null, otherwise

    edited_at: Optional[:py:class:`datetime.datetime`]

    edited_by_id: Optional[int]
        user id of the user that edited the post; null, otherwise

    id: int
        the ID of the comment

    legacy_name: Optional[str]
        username displayed on legacy comments

    message: Optional[str]
        markdown of the comment's content

    message_html: Optional[str]
        html version of the comment's content

    parent_id: Optional[int]
        ID of the comment's parent

    pinned: bool
        Pin status of the comment

    replies_count: int
        number of replies to the comment

    updated_at: :py:class:`datetime.datetime`

    user: Optional[:class:`UserCompact`]

    user_id: int
        user ID of the poster

    votes_count: int
        number of votes

    url: str
        URL to the comment
    """

    __slots__ = (
        "commentable_id",
        "commentable_type",
        "created_at",
        "deleted_at",
        "deleted_by_id",
        "edited_at",
        "edited_by_id",
        "id",
        "legacy_name",
        "message",
        "message_html",
        "parent_id",
        "pinned",
        "replies_count",
        "updated_at",
        "user",
        "user_id",
        "votes_count",
    )

    def __init__(self, data):
        self.commentable_id: int = get_required(data, "commentable_id")
        self.commentable_type: str = get_required(data, "commentable_type")
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", fromisoformat)
        self.deleted_by_id: Optional[int] = data.get("deleted_by_id")
        self.edited_at: Optional[datetime] = get_optional(data, "edited_at", fromisoformat)
        self.edited_by_id: Optional[int] = get_required(data, "edited_by_id")
        self.id: int = get_required(data, "id")
        self.legacy_name: Optional[str] = get_required(data, "legacy_name")
        self.message: Optional[str] = data.get("message")
        self.message_html: Optional[str] = data.get("message_html")
        self.parent_id: Optional[int] = get_required(data, "parent_id")
        self.pinned: bool = get_required(data, "pinned")
        self.replies_count: int = get_required(data, "replies_count")
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))
        self.user: Optional[UserCompact] = get_optional(data, "user", UserCompact)
        self.user_id: int = get_required(data, "user_id")
        self.votes_count: int = get_required(data, "votes_count")

    @property
    def url(self) -> str:
        return f"https://osu.ppy.sh/comments/{self.id}"

    def __repr__(self):
        return prettify(self, "user_id", "message")


class CommentBundle:
    """
    Comments and related data.

    **Attributes**

    commentable_meta: List[:class:`CommentableMeta`]
        ID of the object the comment is attached to

    comments: List[:class:`Comment`]
        List of comments ordered according to sort

    has_more: bool
        If there are more comments or replies available

    has_more_id: int

    included_comments: List[:class:`Comment`]
        Related comments; e.g. parent comments and nested replies

    pinned_comments: List[:class:`Comment`]
        Pinned comments

    sort: :class:`CommentSort`

    cursor: :class:`dict`

    top_level_count: Optional[int]
        Number of comments at the top level. Not returned for replies.

    total: Optional[int]
        Total number of comments. Not returned for replies.

    user_follow: bool
        is the current user watching the comment thread?

    user_votes: List[int]
        IDs of the comments in the bundle the current user has upvoted

    users: List[:class:`UserCompact`]
        List of users related to the comments
    """

    __slots__ = (
        "commentable_meta",
        "comments",
        "has_more",
        "has_more_id",
        "included_comments",
        "pinned_comments",
        "sort",
        "cursor",
        "top_level_count",
        "total",
        "user_follow",
        "user_votes",
        "users",
    )

    def __init__(self, data):
        self.commentable_meta: List[CommentableMeta] = list(
            map(CommentableMeta, get_required(data, "commentable_meta"))
        )
        self.comments: List[Comment] = list(map(Comment, get_required(data, "comments")))
        self.has_more: bool = get_required(data, "has_more")
        self.has_more_id: int = get_required(data, "has_more_id")
        self.included_comments: List[Comment] = list(map(Comment, get_required(data, "included_comments")))
        self.pinned_comments: List[Comment] = list(map(Comment, get_required(data, "pinned_comments")))
        self.sort: CommentSort = CommentSort(get_required(data, "sort"))
        self.cursor: dict = get_required(data, "cursor")
        self.top_level_count: Optional[int] = data.get("top_level_count")
        self.total: Optional[int] = data.get("total")
        self.user_follow: bool = get_required(data, "user_follow")
        self.user_votes: List[int] = get_required(data, "user_votes")
        self.users: List[UserCompact] = list(map(UserCompact, get_required(data, "users")))

    def __repr__(self):
        return prettify(self, "commentable_meta", "comments", "users")


class CommentableMeta:
    """
    Metadata of the object that a comment is attached to.
    If the object is deleted then title is the only attribute with a value.

    **Attributes**

    id: Optional[int]
        the ID of the object

    owner_id: Optional[int]
        the ID of the owner of the object

    owner_title: Optional[int]
        undocumented

    title: Optional[str]
        display title

    type: Optional[str]
        the type of the object

    url: Optional[str]
        url of the object

    current_user_attributes: :class:`CommentableMetaAttributes`
    """

    __slots__ = (
        "id",
        "title",
        "type",
        "url",
        "owner_id",
        "owner_title",
        "current_user_attributes",
    )

    def __init__(self, data):
        self.id: Optional[int] = data.get("id")
        self.owner_id: Optional[int] = data.get("owner_id")
        self.owner_title: Optional[str] = data.get("owner_title")
        self.title: str = data.get("title")
        self.type: Optional[str] = data.get("type")
        self.url: Optional[str] = data.get("url")
        self.current_user_attributes: Optional[CommentableMetaAttributes] = get_optional(
            data, "current_user_attributes", CommentableMetaAttributes
        )

    def __repr__(self):
        return prettify(self, "title", "url")



from .enums import (
    Mods,
    Enum,
    BeatmapsetSearchSort,
    BeatmapsetSearchGeneral,
    BeatmapsetSearchPlayed,
    BeatmapsetSearchStatus,
    BeatmapsetSearchExtra,
    BeatmapsetLanguage,
    BeatmapsetGenre,
    ScoreRank,
    GameModeInt,
    Mod,
    NotificationCategory,
    ObjectType,
)
from typing import Any, Sequence, Union, Optional, TypeVar, List, Callable, Dict
from datetime import datetime, timezone
import os
import re


__all__ = ("BeatmapsetSearchFilter", "PlaylistItemUtil", "NotificationsUtil", "IdentitiesUtil")


_T = TypeVar("_T")
_V = TypeVar("_V")


def check_scope(func):
    def check(self, other):
        if not isinstance(other, self.__class__) and type(other) != str:
            raise TypeError(f"Cannot compare type Scope with type {type(other).__name__}")
        if isinstance(other, self.__class__):
            other = other.scope
        if other not in self.valid_scopes:
            raise NameError(f"{other} is not a valid scope.")
        return func(self, other)

    return check


def create_autoclass_for_sphinx():
    with open("objects.py", "r") as f:
        info = f.readline()
        while info:
            if info.startswith("class"):
                name = info.split()[1]
                if "(" in name:
                    name = name.split("(")[0]
                else:
                    name = name[:-1]
                content = f"{name}\n{'^' * len(name)}\n.. autoclass:: osu.{name}\n   :members:\n\n"
                with open("text.txt", "a") as f2:
                    f2.write(content)
                    f2.close()
            info = f.readline()
        f.close()


def parse_mods_arg(mods):
    if mods is None:
        return
    if isinstance(mods, Mods):
        return mods.value
    if isinstance(mods, Sequence):
        if len(mods) == 0:
            return
        return Mods.parse_any_list(mods).value
    raise TypeError(f"mods argument must be of type Mods or Sequence, not {type(mods)}")


def prettify(cls: object, *fields: str) -> str:
    d = {s: getattr(cls, s) for s in fields}
    return cls.__class__.__qualname__ + "(" + ", ".join([f"{k}={d[k]!r}" for k in d]) + ")"


def parse_enum_args(*args) -> Union[List[Any], Any]:
    args = [arg.value if isinstance(arg, Enum) else arg for arg in args]
    return args if len(args) != 1 else args[0]


def create_multipart_formdata(data: dict):
    return {k: (None, v, "text/plain", {"charset": "utf-8"}) for k, v in data.items()}


class Util:
    @staticmethod
    def int(value):
        if value is None:
            return
        return int(value)

    @staticmethod
    def float(value):
        if value is None:
            return
        return float(value)


class BeatmapsetSearchFilter:
    """
    Util class that helps for filtering in :func:`Client.search_beatmapsets`.

    Read about each filter under its corresponding function.

    All set functions return the instance of the class to allow for chaining.
    """

    def __init__(self):
        self._filters = {
            "query": None,
            "sort": None,
            "nsfw": None,
            "played": None,
            "r": None,
            "m": None,
            "l": None,
            "g": None,
            "e": None,
            "c": None,
            "s": None,
        }

    def set_query(self, query: str) -> "BeatmapsetSearchFilter":
        """
        Set the query to search for.

        query: str
        """
        self._filters["query"] = query
        return self

    def set_sort(self, sort: Union[BeatmapsetSearchSort, str], order: str = "desc") -> "BeatmapsetSearchFilter":
        """
        Set the sort order of the search.

        sort: Union[:class:`BeatmapsetSearchSort`, str]

        order: str
            "desc" or "asc"
        """
        self._filters["sort"] = (sort.value if isinstance(sort, BeatmapsetSearchSort) else sort) + "_" + order
        return self

    def set_nsfw(self, nsfw: bool) -> "BeatmapsetSearchFilter":
        """
        Set whether to include NSFW in the search.

        nsfw: bool
        """
        self._filters["nsfw"] = nsfw
        return self

    def set_played(self, played: Union[BeatmapsetSearchPlayed, str]) -> "BeatmapsetSearchFilter":
        """
        Set whether to include played and unplayed beatmapsets in the search.

        played: Union[:class:`BeatmapsetSearchPlayed`, str]
        """
        self._filters["played"] = played.value if isinstance(played, BeatmapsetSearchPlayed) else played
        return self

    def set_ranked(self, rank: Union[ScoreRank, str]) -> "BeatmapsetSearchFilter":
        """
        Filter by rank achieved.

        rank: Union[:class:`ScoreRank`, str]
        """
        self._filters["r"] = rank.value if isinstance(rank, ScoreRank) else rank
        return self

    def set_mode(self, mode: Union[GameModeInt, int]) -> "BeatmapsetSearchFilter":
        """
        Set the game mode to filter by.

        mode: Union[:class:`GameModeInt`, str]
        """
        self._filters["m"] = mode.value if isinstance(mode, GameModeInt) else mode
        return self

    def set_language(self, language: Union[BeatmapsetLanguage, int]) -> "BeatmapsetSearchFilter":
        """
        Set the language to filter by.

        language: str
        """
        self._filters["l"] = language.value if isinstance(language, BeatmapsetLanguage) else language
        return self

    def set_genre(self, genre: Union[BeatmapsetGenre, int]) -> "BeatmapsetSearchFilter":
        """
        Set the genre to filter by.

        genre: str
        """
        self._filters["g"] = genre.value if isinstance(genre, BeatmapsetGenre) else genre
        return self

    def set_extra(self, extras: Sequence[Union[BeatmapsetSearchExtra, str]]) -> "BeatmapsetSearchFilter":
        """
        Set the extras to filter by.

        extras: Sequence[Union[:class:`BeatmapsetSearchExtra`, str]]
        """
        self._filters["e"] = ".".join(
            list(
                map(
                    lambda x: x.value if isinstance(x, BeatmapsetSearchExtra) else x,
                    extras,
                )
            )
        )
        return self

    def set_generals(self, generals: Sequence[Union[BeatmapsetSearchGeneral, str]]) -> "BeatmapsetSearchFilter":
        """
        Set the generals to filter by.

        generals: Sequence[Union[:class:`BeatmapsetSearchGeneral`, str]]
        """
        self._filters["c"] = ".".join(
            list(
                map(
                    lambda x: x.value if isinstance(x, BeatmapsetSearchGeneral) else x,
                    generals,
                )
            )
        )
        return self

    def set_status(self, status: Union[BeatmapsetSearchStatus, str]) -> "BeatmapsetSearchFilter":
        """
        Set the status to filter by.

        status: Union[:class:`BeatmapsetSearchStatus`, str]
        """
        self._filters["s"] = status.value if isinstance(status, BeatmapsetSearchStatus) else status
        return self

    @property
    def filters(self) -> dict:
        """
        Dictionary of all filters only including ones that have been set.
        """
        return dict(filter(lambda item: item[1] is not None, self._filters.items()))


class PlaylistItemUtil:
    """
    Util class for passing playlist items to endpoints that involve creating a room
    like create_multiplayer_room and create_playlist.

    **Init Parameters**

    beatmap_id: int

    ruleset_id: Union[:class:`GameModeInt`, int]

    allowed_mods: Sequence[:class:`Mod`]

    required_mods: Sequence[:class:`Mod`]

    **Properties**

    json: :class:`dict`
        Dictionary format of all the attributes which the client uses
        when sending a http request
    """

    __slots__ = ("beatmap_id", "ruleset_id", "allowed_mods", "required_mods")

    def __init__(
        self,
        beatmap_id: int,
        ruleset_id: Union[GameModeInt, int],
        allowed_mods: Sequence[Mod] = None,
        required_mods: Sequence[Mod] = None,
    ):
        self.beatmap_id = beatmap_id
        self.ruleset_id = ruleset_id
        self.allowed_mods = allowed_mods
        self.required_mods = required_mods

    @property
    def json(self) -> dict:
        def mod_map(mod):
            return {"acronym": mod.value}

        return {
            "beatmap_id": self.beatmap_id,
            "ruleset_id": parse_enum_args(self.ruleset_id),
            "allowed_mods": list(map(mod_map, self.allowed_mods)) if self.allowed_mods is not None else None,
            "required_mods": list(map(mod_map, self.required_mods)) if self.required_mods is not None else None,
        }


class JsonUtil:
    __slots__ = ()

    def __init__(self, *args):
        for attr, value in zip(self.__slots__, args):
            setattr(self, attr, value)

    def __getitem__(self, item):
        return getattr(self, item)

    def values(self):
        return [getattr(self, attr) for attr in self.__slots__]

    def json(self):
        return {attr: getattr(self, attr) for attr in self.__slots__}

    def items(self):
        return map(lambda attr: (attr, getattr(self, attr)), self.__slots__)

    @staticmethod
    def parse_list(l):
        return list(map(lambda item: item.json if isinstance(item, JsonUtil) else item, l))

    @staticmethod
    def list_to_labeled_dict(l, name):
        return dict(
            sum(
                map(
                    lambda i: list(
                        map(
                            lambda i2: (f"{name}[{i[0]}][{i2[0]}]", f"{i2[1]}"),
                            filter(lambda i2: i2[1] is not None, i[1].items()),
                        )
                    ),
                    enumerate(l),
                ),
                [],
            )
        )


class NotificationsUtil(JsonUtil):
    """
    **Init parameters**

    category: Optional[Union[str, :class:`NotificationCategory`]]
        Notification category.

    id: Optional[int]
        Id of notification to be marked as read

    object_id: Optional[str]
        Id of the object that triggered the notification

    object_type: Optional[Union[str, :class:`ObjectType`]]
        Type of the object that triggered the notification
    """

    __slots__ = ("category", "id", "object_id", "object_type")

    def __init__(
        self,
        category: Optional[Union[NotificationCategory, str]] = None,
        id: Optional[int] = None,
        object_id: Optional[str] = None,
        object_type: Optional[Union[ObjectType, str]] = None,
    ):
        super().__init__(*parse_enum_args(category, id, object_id, object_type))


class IdentitiesUtil(JsonUtil):
    """
    **Init parameters**

    category: Optional[Union[str, :class:`NotificationCategory`]]
        Notification category.

    object_id: Optional[str]
        Id of the object that triggered the notification

    object_type: Optional[Union[str, :class:`ObjectType`]]
        Type of the object that triggered the notification
    """

    __slots__ = ("category", "object_id", "object_type")

    def __init__(
        self,
        category: Optional[Union[NotificationCategory, str]] = None,
        object_id: Optional[str] = None,
        object_type: Optional[Union[ObjectType, str]] = None,
    ):
        super().__init__(*parse_enum_args(category, object_id, object_type))


def get_optional(data: Dict[str, _V], key: str, call: Callable[[_V], _T]) -> Optional[_T]:
    return call(value) if (value := data.get(key)) is not None else None


def get_optional_list(data: Dict[str, _V], key: str, obj: Callable[[_V], _T]) -> Optional[List[_T]]:
    return list(map(obj, value)) if (value := data.get(key)) is not None else None


# if testing, raise an error for expected keys
# otherwise return none to avoid key errors in prod
if os.getenv("OSUPY_TEST") is None:

    def get_required(data: Dict[str, _V], key: str):
        return data.get(key)

else:

    def get_required(data: Dict[str, _V], key: str):
        return data[key]


ISO_RE = re.compile(r"(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})Z")


def fromisoformat(timestamp: str) -> datetime:
    if timestamp is None:
        raise TypeError("timestamp cannot be None")

    try:
        match = ISO_RE.match(timestamp)
        if match is None:
            return datetime.fromisoformat(timestamp)

        year, month, day, hour, minute, second = match.groups()
        return datetime(
            year=int(year),
            month=int(month),
            day=int(day),
            hour=int(hour),
            minute=int(minute),
            second=int(second),
            tzinfo=timezone.utc,
        )

    # just in case
    except ValueError:
        return datetime.fromisoformat(timestamp)



import time
import asyncio
from typing import Optional, List, AsyncGenerator, TYPE_CHECKING
from inspect import iscoroutinefunction

try:
    import aiohttp
except ImportError:
    aiohttp = None

from ..http import BaseHTTPHandler, HTTPHandler
from ..exceptions import RequestException

if TYPE_CHECKING:
    from .auth import BaseAsynchronousAuthHandler
    from ..auth import BaseAuthHandler


__all__ = ("AsynchronousHTTPHandler", "BaseAsynchronousHTTPHandler")


class BaseAsynchronousHTTPHandler(BaseHTTPHandler):
    auth: "BaseAsynchronousAuthHandler"

    def __init__(self, auth: Optional["BaseAsynchronousAuthHandler"], api_version: Optional[str] = None):
        super().__init__(auth, api_version)

        if not iscoroutinefunction(self.auth.get_token):
            raise ValueError("auth must have an async get_token method")

        if aiohttp is None:
            raise RuntimeError(
                "Missing aiohttp package, which is required to use asynchronous features."
                'Install osu.py with the async feature: "pip install osu.py[async]"'
            )

    def get_req_gen(self, path, *args, **kwargs) -> AsyncGenerator:
        raise NotImplementedError()

    async def make_request(self, path, *args, **kwargs):
        raise NotImplementedError()


class AsynchronousHTTPHandler(BaseAsynchronousHTTPHandler):
    """
    Handles making asynchronous requests. Used by :class:`osu.AsynchronousClient`.
    """

    def __init__(
        self,
        auth: Optional["BaseAsynchronousAuthHandler"],
        request_wait_time: float = 1.0,
        limit_per_minute: int = 60,
        api_version: Optional[str] = None,
    ):
        super().__init__(auth, api_version)

        self.rate_limit: RateLimitHandler = RateLimitHandler(request_wait_time, limit_per_minute)

    def set_ratelimit(self, request_wait_time: float = 1.0, limit_per_minute: int = 60):
        self.rate_limit.wait_time = request_wait_time
        self.rate_limit.limit = limit_per_minute

    async def get_headers(self, path, is_files=False, **kwargs):
        headers = {
            "charset": "utf-8",
            "x-api-version": self.api_version,
            "Accept": path.accept,
        }

        if not is_files:  # otherwise let requests library handle it
            headers["Content-Type"] = path.content_type

        if path.requires_auth and "Authorization" not in headers:
            token = await self.auth.get_token()
            if token is None:
                raise ValueError("Cannot make request requiring authorization with a null token")
            headers["Authorization"] = f"Bearer {token}"

        for key, value in kwargs.items():
            if value is not None:
                headers[str(key)] = str(value)

        return headers

    async def make_request_to_endpoint(self, endpoint, path, data=None, headers=None, files=None, **kwargs):
        if headers is None:
            headers = {}
        json = data
        file_data = None

        self.check_path_validity(path)

        headers = await self.get_headers(path, files is not None, **headers)
        params = {str(key): value for key, value in kwargs.items() if value is not None}
        if files is not None:
            file_data = dict(map(lambda item: (item[0], item[1][1]), files.items()))

        async with self.rate_limit:
            async with aiohttp.ClientSession() as session:
                async with session.request(
                    path.method,
                    endpoint + path.path,
                    headers=headers,
                    data=file_data,
                    json=json,
                    params=params,
                ) as resp:
                    await self._raise_for_status(resp)

                    if resp.content_length == 0:
                        return
                    yield resp

    async def _raise_for_status(self, resp):
        try:
            resp.raise_for_status()
        except Exception as e:
            try:
                err = (await resp.json())["error"]
            except:
                err = None

            if err:
                raise RequestException(err) from e

            raise e

    def get_req_gen(self, path, *args, **kwargs):
        return self.make_request_to_endpoint(self.base_url, path, *args, **kwargs)

    async def make_request(self, path, *args, **kwargs):
        """
        Make request to the api.

        :param path:
        :type path: :class:`osu.Path`

        :param data: (Default None)
            Json body of the request
        :type data: Optional[Union[dict, list]]

        :param headers: (Default None)
            Headers to send in the request
        :type headers: Optional[Dict[str, str]]

        :param is_download: (Default False)
            Returns response object if true
        :type is_download: bool

        :param files: (Default None)

        :param kwargs:
            All kwargs will be interpreted as query parameters for the request.
        :type kwargs: Dict[str, str]
        """
        gen = self.get_req_gen(path, *args, **kwargs)
        async for resp in gen:
            try:
                return await resp.json()
            except aiohttp.client_exceptions.ContentTypeError:
                return

    async def make_auth_request(self, data):
        async with self.rate_limit:
            async with aiohttp.ClientSession() as session:
                async with session.request("POST", self.token_url, json=data) as resp:
                    await self._raise_for_status(resp)
                    return await resp.json()

    @classmethod
    def from_sync(cls, http: HTTPHandler, auth: Optional["BaseAsynchronousAuthHandler"] = None):
        new_http = cls(auth, http.rate_limit.wait_time, http.rate_limit.limit, http.api_version)
        new_http.rate_limit._requests_finished = http.rate_limit._requests_finished
        new_http.base_url = http.base_url
        new_http.auth_url = http.auth_url
        new_http.token_url = http.token_url
        return new_http

    def as_sync(self, auth: Optional["BaseAuthHandler"]) -> HTTPHandler:
        return HTTPHandler.from_async(self, auth)


class RateLimitHandler:
    __slots__ = (
        "wait_time",
        "limit",
        "_lock",
        "_waiting_lock",
        "_finish_evt",
        "_requests_in_progress",
        "_requests_finished",
    )

    def __init__(self, request_wait_time: float, limit_per_minute: int):
        self.wait_time: float = request_wait_time
        self.limit: int = limit_per_minute
        # for accessing non-thread-safe variables
        # intended to be used for short durations
        self._lock: asyncio.Lock = asyncio.Lock()
        # to make sure only one request is waiting at a time
        # okay to acquire for long periods of time
        self._waiting_lock: asyncio.Lock = asyncio.Lock()
        self._finish_evt: asyncio.Event = asyncio.Event()
        self._requests_in_progress = 0
        self._requests_finished: List[float] = []

    async def __aenter__(self):
        await self._lock.acquire()

        if self.wait_time > 0:
            await self._wait_with_wait_time()
        else:
            await self._wait_without_wait_time()

        self._requests_in_progress += 1

        self._lock.release()

    async def _wait_with_wait_time(self):
        # acquiring _waiting_lock could take a bit
        # so let's release this one
        self._lock.release()
        # once acquired, we can choose the appropriate way to wait
        # without worry about race conditions. waiting one at a time
        # is perfectly valid since wait time between requests is > 0
        await self._waiting_lock.acquire()

        await self._lock.acquire()
        # wait for current request to finish
        if self._requests_in_progress > 0:
            # clear before releasing to make sure a request doesn't
            # finish in between and cause a deadlock
            self._finish_evt.clear()
            self._lock.release()

            await self._finish_evt.wait()
            time.sleep(self.wait_time)
            await self._lock.acquire()
        # wait until enough time as passed since last request finished
        elif len(requests_finished := self._get_requests_finished()) > 0:
            wait_time = max(0.0, self.wait_time - (time.monotonic() - requests_finished[-1]))
            if wait_time > 0:
                self._lock.release()
                time.sleep(wait_time)
                await self._lock.acquire()

        self._waiting_lock.release()

    async def _wait_without_wait_time(self):
        # under rate limit still
        if self._requests_in_progress + len(self._get_requests_finished()) < self.limit:
            return

        # acquiring _waiting_lock could take a bit
        # so let's release this one
        self._lock.release()
        await self._waiting_lock.acquire()

        await self._lock.acquire()
        should_wait = len(self._get_requests_finished()) == 0
        # if a request finishes between declaring should_wait
        # and the if statement below, it will know because
        # _finish_evt will be set
        self._finish_evt.clear()
        self._lock.release()

        # can't wait based on request history
        # so we wait until a request finishes and sets self._finish_evt
        if should_wait:
            await self._finish_evt.wait()

        await self._lock.acquire()
        oldest_req = None if len(requests_finished := self._get_requests_finished()) == 0 else requests_finished[0]
        should_wait = len(requests_finished) + self._requests_in_progress >= self.limit
        self._lock.release()

        # wait until back under limit
        if should_wait and oldest_req is not None:
            wait_time = max(0.0, 60 - (time.monotonic() - oldest_req))
            if wait_time > 0:
                time.sleep(wait_time)

        self._waiting_lock.release()
        await self._lock.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        finished_at = time.monotonic()
        async with self._lock:
            self._requests_in_progress -= 1
            self._get_requests_finished().append(finished_at)
            # alert any waiting requests that one has just finished
            self._finish_evt.set()

    def _get_requests_finished(self):
        """expects self._lock is acquired when calling this function"""
        # update list
        while len(self._requests_finished) > 0 and time.monotonic() - self._requests_finished[0] >= 60:
            self._requests_finished.pop(0)

        return self._requests_finished



from typing import Optional, List, TYPE_CHECKING, Union, Dict

from .beatmap import BeatmapCompact, BeatmapsetCompact, Beatmap
from .user import UserCompact
from .current_user_attributes import ScoreUserAttributes
from ..enums import GameModeStr, GameModeInt, Mods, Mod, ObjectType, ScoreRank
from ..util import prettify, get_optional, get_required, fromisoformat

if TYPE_CHECKING:
    from datetime import datetime


__all__ = (
    "BeatmapScores",
    "LegacyScore",
    "SoloScore",
    "PpWeight",
    "ScoreStatistics",
    "ScoreDataStatistics",
    "LazerMod",
    "BeatmapUserScore",
    "get_score_object",
)


class BeatmapScores:
    """
    Contains a list of scores as well as, possibly, a :class:`BeatmapUserScore` object.

    **Attributes**

    scores: List[Union[:class:`LegacyScore`, :class:`SoloScore`]]
        The list of top scores for the beatmap in descending order.

    user_score: Optional[:class:`BeatmapUserScore`]
        The score of the current user. This is not returned if the current user does not have a score.
    """

    __slots__ = ("scores", "user_score")

    def __init__(self, data, api_version: str):
        self.scores: List[Union[LegacyScore, SoloScore]] = [
            get_score_object(score, api_version) for score in get_required(data, "scores")
        ]
        # supposedly will be switched to 'user_score' in the future
        var_name = "userScore" if "userScore" in data else "user_score"
        self.user_score: Optional[BeatmapUserScore] = get_optional(data, var_name, BeatmapUserScore)

    def __repr__(self):
        return prettify(self, "user_score", "scores")

    def __len__(self):
        return len(self.scores)

    def __iter__(self):
        return iter(self.scores)


class LegacyScore:
    """
    Contains information about a score

    **Attributes**

    id: int

    best_id: int

    user_id: int

    accuracy: float

    mods: :class:`Mods`

    score: int

    max_combo: int

    perfect: bool

    statistics: :class:`ScoreStatistics`

    passed bool

    pp: float

    rank: :class:`ScoreRank`

    created_at: :py:class:`datetime.datetime`

    mode: :class:`GameModeStr`

    mode_int: :class:`GameModeInt`

    has_replay: bool
        is the replay is available

    beatmap: Optional[:class:`BeatmapCompact`]

    beatmapset: Optional[:class:`BeatmapsetCompact`]

    rank_country: Optional[int]

    rank_global: Optional[int]

    weight: Optional[:class:`PpWeight`]

    user: Optional[:class:`UserCompact`]

    match: Optional[:class:`MatchGameScoreInfo`]

    current_user_attributes: Optional[:class:`ScoreUserAttributes`]
    """

    __slots__ = (
        "id",
        "best_id",
        "user_id",
        "accuracy",
        "mods",
        "score",
        "max_combo",
        "perfect",
        "statistics",
        "passed",
        "pp",
        "rank",
        "created_at",
        "mode",
        "mode_int",
        "has_replay",
        "beatmap",
        "beatmapset",
        "rank_country",
        "rank_global",
        "weight",
        "user",
        "match",
        "type",
        "current_user_attributes",
    )

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.best_id: int = get_required(data, "best_id")
        self.user_id: int = get_required(data, "user_id")
        self.accuracy: float = get_required(data, "accuracy")
        self.mods: Mods = Mods.parse_any_list(get_required(data, "mods"))
        self.score: int = get_required(data, "score")
        self.max_combo: int = get_required(data, "max_combo")
        self.perfect: bool = get_required(data, "perfect")
        self.statistics: ScoreStatistics = ScoreStatistics(get_required(data, "statistics"))
        self.passed: bool = get_required(data, "passed")
        self.pp: float = get_required(data, "pp")
        self.rank: ScoreRank = ScoreRank(get_required(data, "rank"))
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.mode: GameModeStr = GameModeStr(get_required(data, "mode"))
        self.mode_int: GameModeInt = GameModeInt(get_required(data, "mode_int"))
        self.has_replay: bool = get_required(data, "replay")
        self.type: ObjectType = ObjectType(get_required(data, "type"))

        from .match import MatchGameScoreInfo

        self.beatmap: Optional[BeatmapCompact] = get_optional(data, "beatmap", BeatmapCompact)
        self.beatmapset: Optional[BeatmapsetCompact] = get_optional(data, "beatmapset", BeatmapsetCompact)
        self.user: Optional[UserCompact] = get_optional(data, "user", UserCompact)
        self.match: Optional[MatchGameScoreInfo] = get_optional(data, "match", MatchGameScoreInfo)
        self.rank_country: Optional[int] = data.get("rank_country")
        self.rank_global: Optional[int] = data.get("rank_global")
        self.weight: Optional[PpWeight] = get_optional(data, "weight", PpWeight)
        self.current_user_attributes: Optional[ScoreUserAttributes] = get_optional(
            data, "current_user_attributes", ScoreUserAttributes
        )

    # backwards compatibility
    @property
    def replay(self):
        return self.has_replay

    def __repr__(self):
        return prettify(self, "user_id", "accuracy")


class SoloScore:
    """
    Contains information about a score in lazer format (may be scores set on stable).

    **Attributes**

    accuracy: float

    beatmap_id: Optional[int]

    beatmap: Optional[:class:`Beatmap`]

    beatmapset: Optional[:class:`BeatmapsetCompact`]

    ended_at: :py:class:`datetime.datetime`

    max_combo: int

    maximum_statistics: :class:`ScoreDataStatistics`

    mods: List[:class:`LazerMod`]

    passed: bool

    rank: :class:`ScoreRank`

    ruleset_id: int

    statistics: :class:`ScoreDataStatistics`

    total_score: int

    user_id: int

    best_id: Optional[int]

    id: int

    legacy_perfect: Optional[bool]

    pp: Optional[float]

    replay: bool

    type: :class:`ObjectType`

    user: Optional[:class:`UserCompact`]

    build_id: Optional[int]

    legacy_score_id: Optional[int]

    legacy_total_score: Optional[int]

    started_at: Optional[:py:class:`datetime.datetime`]

    current_user_attributes: Optional[:class:`ScoreUserAttributes`]

    weight: Optional[:class:`PpWeight`]
    """

    __slots__ = (
        "accuracy",
        "beatmap_id",
        "build_id",
        "ended_at",
        "legacy_score_id",
        "legacy_total_score",
        "max_combo",
        "maximum_statistics",
        "mods",
        "passed",
        "rank",
        "ruleset_id",
        "started_at",
        "statistics",
        "total_score",
        "user_id",
        "best_id",
        "id",
        "legacy_perfect",
        "pp",
        "replay",
        "type",
        "user",
        "current_user_attributes",
        "weight",
        "beatmap",
        "beatmapset",
    )

    def __init__(self, data):
        self.accuracy: float = get_required(data, "accuracy")
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.ended_at: datetime = fromisoformat(get_required(data, "ended_at"))
        self.max_combo: int = get_required(data, "max_combo")
        self.maximum_statistics: ScoreDataStatistics = ScoreDataStatistics(get_required(data, "maximum_statistics"))
        self.mods: List[LazerMod] = list(map(LazerMod, get_required(data, "mods")))
        self.passed: bool = get_required(data, "passed")
        self.rank: ScoreRank = ScoreRank(get_required(data, "rank"))
        self.ruleset_id: int = get_required(data, "ruleset_id")
        self.statistics: ScoreDataStatistics = ScoreDataStatistics(get_required(data, "statistics"))
        self.total_score: int = get_required(data, "total_score")
        self.user_id: int = get_required(data, "user_id")
        self.best_id: Optional[int] = get_required(data, "best_id")
        self.id: int = get_required(data, "id")
        self.legacy_perfect: Optional[bool] = get_required(data, "legacy_perfect")
        self.pp: Optional[float] = get_required(data, "pp")
        self.replay: bool = get_required(data, "replay")
        self.type: ObjectType = ObjectType(get_required(data, "type"))

        self.user: Optional[UserCompact] = get_optional(data, "user", UserCompact)
        self.build_id: Optional[int] = data.get("build_id")
        self.legacy_score_id: Optional[int] = data.get("legacy_score_id")
        self.legacy_total_score: Optional[int] = data.get("legacy_total_score")
        self.started_at: Optional[datetime] = get_optional(data, "started_at", fromisoformat)
        self.current_user_attributes: Optional[ScoreUserAttributes] = get_optional(
            data, "current_user_attributes", ScoreUserAttributes
        )
        self.weight: Optional[PpWeight] = get_optional(data, "weight", PpWeight)
        self.beatmap: Optional[Beatmap] = get_optional(data, "beatmap", Beatmap)
        self.beatmapset: Optional[BeatmapsetCompact] = get_optional(data, "beatmapset", BeatmapsetCompact)

    def __repr__(self):
        return prettify(self, "beatmap_id", "statistics")


class PpWeight:
    """
    Weighted pp info

    **Attributes**

    percentage: int
        number (0-100) that tells percentage weighed

    pp: float
        amount of pp after being weighted
    """

    __slots__ = ("percentage", "pp")

    def __init__(self, data):
        self.percentage: float = get_required(data, "percentage")
        self.pp: float = get_required(data, "pp")


def get_score_object(data, api_version: str) -> Union[SoloScore, LegacyScore]:
    if int(api_version) > 20220704:
        return SoloScore(data)

    return LegacyScore(data)


class ScoreStatistics:
    """
    **Attributes**

    count_50: int

    count_100: int

    count_300: int

    count_geki: int

    count_katu: int

    count_miss: int
    """

    __slots__ = (
        "count_50",
        "count_100",
        "count_300",
        "count_geki",
        "count_katu",
        "count_miss",
    )

    def __init__(self, data):
        self.count_50: int = get_required(data, "count_50")
        self.count_100: int = get_required(data, "count_100")
        self.count_300: int = get_required(data, "count_300")
        self.count_geki: int = get_required(data, "count_geki")
        self.count_katu: int = get_required(data, "count_katu")
        self.count_miss: int = get_required(data, "count_miss")

    def __repr__(self):
        return prettify(self, "count_300", "count_miss")


class ScoreDataStatistics:
    """
    **Attributes**

    ok: Optional[int]

    meh: Optional[int]

    good: Optional[int]

    miss: Optional[int]

    none: Optional[int]

    great: Optional[int]

    perfect: Optional[int]

    ignore_hit: Optional[int]

    ignore_miss: Optional[int]

    large_bonus: Optional[int]

    small_bonus: Optional[int]

    large_tick_hit: Optional[int]

    small_tick_hit: Optional[int]

    large_tick_miss: Optional[int]

    small_tick_miss: Optional[int]

    slider_tail_hit: Optional[int]

    combo_break: Optional[int]

    legacy_combo_increase: Optional[int]
    """

    __slots__ = (
        "ok",
        "meh",
        "good",
        "miss",
        "none",
        "great",
        "perfect",
        "ignore_hit",
        "ignore_miss",
        "large_bonus",
        "small_bonus",
        "large_tick_hit",
        "small_tick_hit",
        "large_tick_miss",
        "small_tick_miss",
        "slider_tail_hit",
        "combo_break",
        "legacy_combo_increase",
    )

    def __init__(self, data):
        self.ok: Optional[int] = data.get("ok")
        self.meh: Optional[int] = data.get("meh")
        self.good: Optional[int] = data.get("good")
        self.miss: Optional[int] = data.get("miss")
        self.none: Optional[int] = data.get("none")
        self.great: Optional[int] = data.get("great")
        self.perfect: Optional[int] = data.get("perfect")
        self.ignore_hit: Optional[int] = data.get("ignore_hit")
        self.ignore_miss: Optional[int] = data.get("ignore_miss")
        self.large_bonus: Optional[int] = data.get("large_bonus")
        self.small_bonus: Optional[int] = data.get("small_bonus")
        self.large_tick_hit: Optional[int] = data.get("large_tick_hit")
        self.small_tick_hit: Optional[int] = data.get("small_tick_hit")
        self.large_tick_miss: Optional[int] = data.get("large_tick_miss")
        self.small_tick_miss: Optional[int] = data.get("small_tick_miss")
        self.slider_tail_hit: Optional[int] = data.get("slider_tail_hit")
        self.combo_break: Optional[int] = data.get("combo_break")
        self.legacy_combo_increase: Optional[int] = data.get("legacy_combo_increase")

    def __repr__(self):
        attrs = tuple(filter(lambda attr: getattr(self, attr) is not None, self.__slots__))
        return prettify(self, *attrs[:2])


class LazerMod:
    """
    **Attributes**

    mod: :class:`Mod`

    settings: Optional[Dict]
    """

    __slots__ = ("mod", "settings")

    def __init__(self, data):
        self.mod: Union[Mod] = Mod(get_required(data, "acronym"))
        self.settings: Optional[Dict] = data.get("settings")

    def __repr__(self):
        return prettify(self, "mod", "settings")


class BeatmapUserScore:
    """
    **Attributes**

    position: int
        The position of the score within the requested beatmap ranking.

    score: :class:`LegacyScore`
        The details of the score.
    """

    __slots__ = ("position", "score")

    def __init__(self, data, api_version: str):
        self.position: int = get_required(data, "position")
        self.score: Union[LegacyScore, SoloScore] = get_score_object(get_required(data, "score"), api_version)

    def __repr__(self):
        return prettify(self, "position")



from typing import TYPE_CHECKING, List

from .user import UserCompact
from ..util import prettify, get_required, fromisoformat

if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("SeasonalBackgrounds", "SeasonalBackground")


class SeasonalBackgrounds:
    """
    Contains data on the seasonal backgrounds.

    **Attributes**

    ends_at: :py:class:`datetime.datetime`
        The date when the seasonal backgrounds will end.

    backgrounds: List[:class:`SeasonalBackground`]
        A list of all the seasonal backgrounds.
    """

    __slots__ = ("ends_at", "backgrounds")

    def __init__(self, data):
        self.ends_at: datetime = fromisoformat(get_required(data, "ends_at"))
        self.backgrounds: List[SeasonalBackground] = list(map(SeasonalBackground, get_required(data, "backgrounds")))

    def __repr__(self):
        return prettify(self, "ends_at", "backgrounds")


class SeasonalBackground:
    """
    Represents a seasonal background.

    **Attributes**

    url: str
        The url of the background.

    user: :class:`UserCompact`
        The artist of the background.
    """

    __slots__ = ("url", "user")

    def __init__(self, data):
        self.url: str = get_required(data, "url")
        self.user: UserCompact = UserCompact(get_required(data, "user"))

    def __repr__(self):
        return prettify(self, "url", "user")



from typing import TYPE_CHECKING, List, Optional, NamedTuple
import math
from collections import namedtuple

from .group import UserGroup
from .forum import TextFormat
from ..util import prettify, get_optional, get_optional_list, get_required, fromisoformat
from ..enums import GameModeStr, GameModeInt, UserAccountHistoryType, UserRelationType


if TYPE_CHECKING:
    from datetime import datetime, date


__all__ = (
    "UserCompact",
    "User",
    "UserPreferences",
    "UserRelation",
    "ProfileBanner",
    "UserSilence",
    "UserAccountHistory",
    "UserBadge",
    "UserMonthlyPlaycount",
    "UserStatistics",
    "UserStatisticVariant",
    "UserStatisticsRulesets",
    "RankHighest",
    "UserAchievement",
    "UserReplaysWatchedCount",
    "RankHistory",
    "UserCover",
    "Country",
    "UserKudosu",
    "DailyChallengeUserStats",
    "UserTeam",
    "UserTeamStatistics",
)


class UserCompact:
    """
    Mainly used for embedding in certain responses to save additional api lookups.

    **Attributes**

    avatar_url: str
        url of user's avatar

    country_code: str
        two-letter code representing user's country

    default_group: str
        Identifier of the default Group the user belongs to.

    id: int
        unique identifier for user

    is_active: bool
        has this account been active in the last x months?

    is_bot: bool
        is this a bot account?

    is_deleted: bool

    is_online: bool
        is the user currently online? (either on lazer or the new website)

    is_supporter: bool
        does this user have supporter?

    last_visit: Optional[:py:class:`datetime.datetime`]
        null if the user hides online presence

    pm_friends_only: bool
        whether or not the user allows PM from other than friends

    profile_colour: Optional[str]
        colour of username/profile highlight, hex code (e.g. #333333)

    username: str
        user's display name

    account_history: Optional[List[:class:`UserAccountHistory`]]

    active_tournament_banner: Optional[:class:`ProfileBanner`]
        deprecated; use active_tournament_banners

    active_tournament_banners: Optional[List[:class:`ProfileBanner`]]

    badges: Optional[List[:class:`UserBadge`]]

    beatmap_playcounts_count: Optional[int]

    blocks: Optional[List[:class:`UserRelations`]]

    comments_count: Optional[int]

    country: Optional[:class:`Country`]

    cover: Optional[:class:`UserCover`]

    daily_challenge_user_stats: Optional[:class:`DailyChallengeUserStats`]

    favourite_beatmapset_count: Optional[int]

    follow_user_mapping: Optional[List[int]]

    follower_count: Optional[int]

    friends: Optional[List[:class:`UserRelation`]]

    graveyard_beatmapset_count: Optional[int]

    groups: Optional[List[:class:`UserGroup`]]

    guest_beatmapset_count: Optional[int]

    is_admin: Optional[bool]

    is_bng: Optional[bool]

    is_gmt: Optional[bool]

    is_limited_bn: Optional[bool]

    is_moderator: Optional[bool]

    is_nat: Optional[bool]

    is_restricted: Optional[bool]

    is_silenced: Optional[bool]

    loved_beatmapset_count: Optional[int]

    mapping_follower_count: Optional[int]

    monthly_playcounts: Optional[List[:class:`UserMonthlyPlaycount`]]

    nominated_beatmapset_count: Optional[int]

    page: Optional[:class:`TextFormat`]

    pending_beatmapset_count: Optional[int]

    previous_usernames: Optional[List[str]]

    rank_highest: Optional[:class:`RankHighest`]

    rank_history: Optional[:class:`RankHistory`]

    ranked_beatmapset_count: Optional[int]

    replays_watched_counts: Optional[List[:class:`UserReplaysWatchedCount`]]

    scores_best_count: Optional[int]

    scores_first_count: Optional[int]

    scores_pinned_count: Optional[int]

    scores_recent_count: Optional[int]

    statistics: Optional[:class:`UserStatistics`]

    statistics_rulesets: Optional[:class:`UserStatisticsRulesets`]

    support_level: Optional[int]

    team: Optional[:class:`UserTeam`]

    unread_pm_count: Optional[int]

    user_achievements: Optional[List[:class:`UserAchievement`]]

    user_preferences: Optional[:class:`UserPreferences`]
    """

    __slots__ = (
        "avatar_url",
        "country_code",
        "default_group",
        "id",
        "is_active",
        "is_bot",
        "is_deleted",
        "is_online",
        "is_supporter",
        "last_visit",
        "pm_friends_only",
        "profile_colour",
        "username",
        "account_history",
        "active_tournament_banner",
        "active_tournament_banners",
        "badges",
        "beatmap_playcounts_count",
        "blocks",
        "comments_count",
        "country",
        "cover",
        "daily_challenge_user_stats",
        "favourite_beatmapset_count",
        "follow_user_mapping",
        "follower_count",
        "friends",
        "graveyard_beatmapset_count",
        "groups",
        "guest_beatmapset_count",
        "is_admin",
        "is_bng",
        "is_full_bn",
        "is_gmt",
        "is_limited_bn",
        "is_moderator",
        "is_nat",
        "is_restricted",
        "is_silenced",
        "loved_beatmapset_count",
        "mapping_follower_count",
        "monthly_playcounts",
        "nominated_beatmapset_count",
        "page",
        "pending_beatmapset_count",
        "previous_usernames",
        "rank_highest",
        "rank_history",
        "ranked_beatmapset_count",
        "replays_watched_counts",
        "scores_best_count",
        "scores_first_count",
        "scores_pinned_count",
        "scores_recent_count",
        "statistics",
        "statistics_rulesets",
        "support_level",
        "team",
        "unread_pm_count",
        "user_achievements",
        "user_preferences",
    )

    def __init__(self, data):
        self.avatar_url: str = get_required(data, "avatar_url")
        self.country_code: str = get_required(data, "country_code")
        self.default_group: str = get_required(data, "default_group")
        self.id: int = get_required(data, "id")
        self.is_active: bool = get_required(data, "is_active")
        self.is_bot: bool = get_required(data, "is_bot")
        self.is_deleted: bool = get_required(data, "is_deleted")
        self.is_online: bool = get_required(data, "is_online")
        self.is_supporter: bool = get_required(data, "is_supporter")
        self.last_visit: Optional[datetime] = get_optional(data, "last_visit", fromisoformat)
        self.pm_friends_only: bool = get_required(data, "pm_friends_only")
        self.profile_colour: Optional[str] = get_required(data, "profile_colour")
        self.username: str = get_required(data, "username")

        # Optional attributes
        self.account_history: Optional[List[UserAccountHistory]] = get_optional_list(
            data, "account_history", UserAccountHistory
        )
        self.active_tournament_banner: Optional[ProfileBanner] = get_optional(
            data, "active_tournament_banner", ProfileBanner
        )
        self.active_tournament_banners: Optional[List[ProfileBanner]] = get_optional_list(
            data, "active_tournament_banners", ProfileBanner
        )
        self.badges: Optional[List[UserBadge]] = get_optional_list(data, "badges", UserBadge)
        self.beatmap_playcounts_count: Optional[int] = data.get("beatmap_playcounts_count")
        self.blocks: Optional[List[UserRelation]] = get_optional_list(data, "blocks", UserRelation)
        self.comments_count: Optional[int] = data.get("comments_count")
        self.country: Optional[Country] = get_optional(data, "country", Country)
        self.cover: Optional[UserCover] = get_optional(data, "cover", UserCover)
        self.daily_challenge_user_stats: Optional[DailyChallengeUserStats] = get_optional(
            data, "daily_challenge_user_stats", DailyChallengeUserStats
        )
        self.favourite_beatmapset_count: Optional[int] = data.get("favourite_beatmapset_count")
        self.follow_user_mapping: Optional[List[int]] = data.get("follower_user_mapping")
        self.follower_count: Optional[int] = data.get("follower_count")
        self.friends: Optional[List[UserRelation]] = get_optional_list(data, "friends", UserRelation)
        self.graveyard_beatmapset_count: Optional[int] = data.get("graveyard_beatmapset_count")
        self.groups: Optional[List[UserGroup]] = get_optional_list(data, "groups", UserGroup)
        self.guest_beatmapset_count: Optional[int] = data.get("guest_beatmapset_count")
        self.is_admin: Optional[bool] = data.get("is_admin")
        self.is_bng: Optional[bool] = data.get("is_bng")
        self.is_full_bn: Optional[bool] = data.get("is_full_bn")
        self.is_gmt: Optional[bool] = data.get("is_gmt")
        self.is_limited_bn: Optional[bool] = data.get("is_limited_bn")
        self.is_moderator: Optional[bool] = data.get("is_moderator")
        self.is_nat: Optional[bool] = data.get("is_nat")
        self.is_restricted: Optional[bool] = data.get("is_restricted")
        self.is_silenced: Optional[bool] = data.get("is_silenced")
        self.loved_beatmapset_count: Optional[int] = data.get("loved_beatmapset_count")
        self.mapping_follower_count: Optional[int] = data.get("mapping_follower_count")
        self.monthly_playcounts: Optional[List[UserMonthlyPlaycount]] = get_optional_list(
            data, "monthly_playcounts", UserMonthlyPlaycount
        )
        self.nominated_beatmapset_count: Optional[int] = data.get("nominated_beatmapset_count")
        self.page: Optional[TextFormat] = get_optional(data, "page", TextFormat)
        self.pending_beatmapset_count: Optional[int] = data.get("pending_beatmapset_count")
        self.previous_usernames: Optional[List[str]] = data.get("previous_usernames")
        self.rank_highest: Optional[RankHighest] = get_optional(data, "rank_highest", RankHighest)
        self.rank_history: Optional[RankHistory] = get_optional(data, "rank_history", RankHistory)
        self.ranked_beatmapset_count: Optional[int] = data.get("ranked_beatmapset_count")
        self.replays_watched_counts: Optional[List[UserReplaysWatchedCount]] = get_optional_list(
            data, "replays_watched_counts", UserReplaysWatchedCount
        )
        self.scores_best_count: Optional[int] = data.get("scores_best_count")
        self.scores_first_count: Optional[int] = data.get("scores_first_count")
        self.scores_pinned_count: Optional[int] = data.get("scores_pinned_count")
        self.scores_recent_count: Optional[int] = data.get("scores_recent_count")
        self.statistics: Optional[UserStatistics] = get_optional(data, "statistics", UserStatistics)
        self.statistics_rulesets: Optional[UserStatisticsRulesets] = get_optional(
            data, "statistics_rulesets", UserStatisticsRulesets
        )
        self.support_level: Optional[int] = data.get("support_level")
        self.team: Optional[UserTeam] = get_optional(data, "team", UserTeam)
        self.unread_pm_count: Optional[int] = data.get("unread_pm_count")
        self.user_achievements: Optional[List[UserAchievement]] = get_optional_list(
            data, "user_achievements", UserAchievement
        )
        self.user_preferences: Optional[UserPreferences] = get_optional(data, "user_preferences", UserPreferences)

    def __repr__(self):
        return prettify(self, "username", "id")


class User(UserCompact):
    """
    Represents a User. Extends :class:`UserCompact` with additional attributes.
    Includes `country`, `cover`, and `is_restricted` attributes of :class:`UserCompact`.

    **Attributes**

    cover_url: str
        url of profile cover. Deprecated, use cover.url instead.

    discord: Optional[str]

    has_supported: bool
        Has been a supporter in the past

    interests: Optional[str]

    join_date: :py:class:`datetime.datetime`

    kudosu: :class:`UserKudosu`

    location: Optional[str]

    max_blocks: int
        maximum number of users allowed to be blocked

    max_friends: int
        maximum number of friends allowed to be added

    occupation: Optional[str]

    playmode: :class:`GameModeStr`

    playstyle: List[str]
        Device choices of the user.

    post_count: int
        number of forum posts

    profile_hue: Optional[int]
        Custom color hue in HSL degrees (0-359).

    profile_order: List[str]
        Ordered list of sections in user profile page. Sections consist of:
        me, recent_activity, beatmaps, historical, kudosu, top_ranks, medals

    title: Optional[str]
        user-specific title

    title_url:  Optional[str]

    twitter:  Optional[str]

    website:  Optional[str]
    """

    __slots__ = (
        "cover_url",
        "discord",
        "has_supported",
        "interests",
        "join_date",
        "kudosu",
        "location",
        "max_blocks",
        "max_friends",
        "occupation",
        "playmode",
        "playstyle",
        "post_count",
        "profile_hue",
        "profile_order",
        "title",
        "title_url",
        "twitter",
        "website",
    )

    def __init__(self, data):
        super().__init__(data)
        self.cover_url: str = data.get("cover_url")
        self.discord: Optional[str] = get_required(data, "discord")
        self.has_supported: bool = get_required(data, "has_supported")
        self.interests: Optional[str] = get_required(data, "interests")
        self.join_date: datetime = fromisoformat(get_required(data, "join_date"))
        self.kudosu: UserKudosu = UserKudosu(get_required(data, "kudosu"))
        self.location: Optional[str] = get_required(data, "location")
        self.max_blocks: int = get_required(data, "max_blocks")
        self.max_friends: int = get_required(data, "max_friends")
        self.occupation: Optional[str] = get_required(data, "occupation")
        self.playmode: GameModeStr = GameModeStr(get_required(data, "playmode"))
        self.playstyle: List[str] = get_required(data, "playstyle")
        self.post_count: int = get_required(data, "post_count")
        self.profile_hue: Optional[int] = get_required(data, "profile_hue")
        self.profile_order: List[str] = get_required(data, "profile_order")
        self.title: Optional[str] = get_required(data, "title")
        self.title_url: Optional[str] = get_required(data, "title_url")
        self.twitter: Optional[str] = get_required(data, "twitter")
        self.website: Optional[str] = get_required(data, "website")


class UserPreferences:
    """
    The settings preferences of a user

    audio_autoplay: bool

    audio_muted: bool

    audio_volume: float

    beatmapset_card_size: str
        normal or extra

    beatmapset_download: str
        all, no_video, or direct

    beatmapset_show_nsfw: bool

    beatmapset_title_show_original: bool

    comments_show_deleted: bool

    forum_posts_show_deleted: bool

    profile_cover_expanded: bool

    user_list_filter: str
        all, online, or offline

    user_list_sort: str
        last_visit, rank, or username

    user_list_view: str
        brick, card, or list
    """

    __slots__ = (
        "audio_autoplay",
        "audio_muted",
        "audio_volume",
        "beatmapset_card_size",
        "beatmapset_download",
        "beatmapset_show_nsfw",
        "beatmapset_title_show_original",
        "comments_show_deleted",
        "forum_posts_show_deleted",
        "profile_cover_expanded",
        "user_list_filter",
        "user_list_sort",
        "user_list_view",
    )

    def __init__(self, data):
        self.audio_autoplay: bool = get_required(data, "audio_autoplay")
        self.audio_muted: bool = get_required(data, "audio_muted")
        self.audio_volume: float = get_required(data, "audio_volume")
        self.beatmapset_card_size: str = get_required(data, "beatmapset_card_size")
        self.beatmapset_download: str = get_required(data, "beatmapset_download")
        self.beatmapset_show_nsfw: bool = get_required(data, "beatmapset_show_nsfw")
        self.beatmapset_title_show_original: bool = get_required(data, "beatmapset_title_show_original")
        self.comments_show_deleted: bool = get_required(data, "comments_show_deleted")
        self.forum_posts_show_deleted: bool = get_required(data, "forum_posts_show_deleted")
        self.profile_cover_expanded: bool = get_required(data, "profile_cover_expanded")
        self.user_list_filter: str = get_required(data, "user_list_filter")
        self.user_list_sort: str = get_required(data, "user_list_sort")
        self.user_list_view: str = get_required(data, "user_list_view")

    def __repr__(self):
        return prettify(self)


class UserRelation:
    """
    Info about relationship to a user

    **Attributes**

    target_id: int
        zebra id

    relation_type: :class:`UserRelationType`

    mutual: bool

    target: Optional[:class:`UserCompact`]
    """

    __slots__ = ("target_id", "relation_type", "mutual", "target")

    def __init__(self, data):
        self.target_id: int = get_required(data, "target_id")
        self.relation_type: UserRelationType = UserRelationType(get_required(data, "relation_type"))
        self.mutual: bool = get_required(data, "mutual")
        self.target: Optional[UserCompact] = get_optional(data, "target", UserCompact)

    def __repr__(self):
        return prettify(self, "target_id" if self.target is None else "target", "mutual")


# legacy support for the old name of the class
UserRelations = UserRelation


class ProfileBanner:
    """
    **Attributes**

    id: int

    tournament_id: int

    image: Optional[str]

    image2x: Optional[str]
    """

    __slots__ = ("id", "tournament_id", "image", "image2x")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.tournament_id: int = get_required(data, "tournament_id")
        self.image: Optional[str] = get_required(data, "image")
        self.image2x: Optional[str] = data["image@2x"]

    def __repr__(self):
        return prettify(self, "tournament_id")


class UserSilence:
    """
    A record indicating a :class:`User` was silenced.

    **Attributes**

    id: int
        id of this object.

    user_id: int
        id of the User that was silenced
    """

    __slots__ = ("id", "user_id")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.user_id: int = get_required(data, "user_id")

    def __repr__(self):
        return prettify(self, "user_id")


class UserAccountHistory:
    """
    **Attributes**

    actor: Optional[:class:`UserCompact`]

    description: str

    id: int

    length: int

    permanent: bool

    supporting_url: Optional[str]

    timestamp: :py:class:`datetime.datetime`

    type: :class:`UserAccountHistoryType`
    """

    __slots__ = (
        "actor",
        "description",
        "id",
        "length",
        "permanent",
        "supporting_url",
        "timestamp",
        "type",
    )

    def __init__(self, data):
        self.actor: Optional[UserCompact] = get_optional(data, "actor", UserCompact)
        self.description: str = get_required(data, "description")
        self.id: int = get_required(data, "id")
        self.length: int = get_required(data, "length")
        self.permanent: bool = get_required(data, "permanent")
        self.supporting_url: Optional[str] = data.get("supporting_url")
        self.timestamp: datetime = fromisoformat(get_required(data, "timestamp"))
        self.type: UserAccountHistoryType = UserAccountHistoryType(get_required(data, "type"))

    def __repr__(self):
        return prettify(self, "type", "length")


class UserBadge:
    """
    **Attributes**

    awarded_at: :py:class:`datetime.datetime`

    description: str

    image_url: str

    image_2x_url: str

    url: str
    """

    __slots__ = ("awarded_at", "description", "image_url", "image_2x_url", "url")

    def __init__(self, data):
        self.awarded_at: datetime = fromisoformat(get_required(data, "awarded_at"))
        self.description: str = get_required(data, "description")
        self.image_url: str = get_required(data, "image_url")
        self.image_2x_url = data["image@2x_url"]
        self.url: str = get_required(data, "url")

    def __repr__(self):
        return prettify(self, "awarded_at")


class UserMonthlyPlaycount:
    """
    **Attributes**

    start_date: :class:`datetime.date`
        year-month-day format

    count: class:`int`
        playcount
    """

    __slots__ = ("start_date", "count")

    def __init__(self, data):
        self.start_date: date = fromisoformat(get_required(data, "start_date")).date()
        self.count: int = get_required(data, "count")

    def __repr__(self):
        return prettify(self, "start_date", "count")


class UserStatistics:
    """
    A summary of various gameplay statistics for a User. Specific to a :class:`GameMode`

    **Attributes**

    count_300: int

    count_100: int

    count_50: int

    count_miss: int

    country_rank: Optional[int]
        Current country rank according to pp.

     global_rank: Optional[int]
        Current global rank according to pp.

    global_rank_exp: Optional[int]
        Current global rank according to experimental pp (not used anymore).

    grade_counts: :class:`NamedTuple`
        Below are the attributes and their meanings.

        a: int
            Number of A ranked scores.

        s: int
            Number of S ranked scores.

        sh: int
            Number of Silver S ranked scores.

        ss: int
            Number of SS ranked scores.

        ssh: int
            Number of Silver SS ranked scores.

    hit_accuracy: float
        Hit accuracy percentage

    is_ranked: bool
        Does the player have a rank

    level: :class:`NamedTuple`
        Has attributes 'current' (current level) and 'progress' (progress to next level).

    maximum_combo: int
        Highest maximum combo.

    play_count: int
        Number of maps played.

    play_time: int
        Cumulative time played.

    pp: int
        Performance points

    pp_exp: Optional[int]
        Experimental performance points (not used anymore)

    rank_change_since_30_days: Optional[int]

    recommended_difficulty: Optional[float]
        Recommended difficulty for a player. This value is not received from the api, but locally calculated.
        The formula is pp^0.4 * 0.195

    recommended_difficulty_exp: Optional[float]
        Recommended difficulty based on the pp_exp value.

    ranked_score: int
        Current ranked score.

    replays_watched_by_others: int
        Number of replays watched by other users.

    total_hits: int
        Total number of hits.

    total_score: int
        Total score.

    user: Optional[:class:`UserCompact`]
        The associated user.

    variants: Optional[List[:class:`UserStatisticVariant`]]
    """

    __slots__ = (
        "grade_counts",
        "level",
        "hit_accuracy",
        "is_ranked",
        "maximum_combo",
        "play_count",
        "play_time",
        "pp",
        "global_rank",
        "ranked_score",
        "replays_watched_by_others",
        "total_hits",
        "total_score",
        "user",
        "country_rank",
        "global_rank_exp",
        "pp_exp",
        "count_100",
        "count_300",
        "count_50",
        "count_miss",
        "recommended_difficulty",
        "recommended_difficulty_exp",
        "variants",
        "rank_change_since_30_days",
    )

    def __init__(self, data):
        self.count_100: int = get_required(data, "count_100")
        self.count_300: int = get_required(data, "count_300")
        self.count_50: int = get_required(data, "count_50")
        self.count_miss: int = get_required(data, "count_miss")
        self.country_rank: Optional[int] = data.get("country_rank")
        self.global_rank: Optional[int] = get_required(data, "global_rank")
        self.global_rank_exp: Optional[int] = data.get("global_rank_exp")
        self.grade_counts: NamedTuple = namedtuple("GradeCounts", ("ssh", "ss", "sh", "s", "a"))(
            **get_required(data, "grade_counts")
        )
        self.level: NamedTuple = namedtuple("Level", ("current", "progress"))(**get_required(data, "level"))
        self.hit_accuracy: float = get_required(data, "hit_accuracy")
        self.is_ranked: bool = get_required(data, "is_ranked")
        self.maximum_combo: int = get_required(data, "maximum_combo")
        self.play_count: int = get_required(data, "play_count")
        self.play_time: int = get_required(data, "play_time")
        self.pp: int = get_required(data, "pp")
        self.pp_exp: Optional[int] = data.get("pp_exp")
        self.recommended_difficulty: Optional[float] = self.calculate_recommended_difficulty(self.pp)
        self.recommended_difficulty_exp: Optional[float] = self.calculate_recommended_difficulty(self.pp_exp)
        self.ranked_score: int = get_required(data, "ranked_score")
        self.replays_watched_by_others: int = get_required(data, "replays_watched_by_others")
        self.total_hits: int = get_required(data, "total_hits")
        self.total_score: int = get_required(data, "total_score")
        self.user: Optional[UserCompact] = get_optional(data, "user", UserCompact)
        self.variants: Optional[List[UserStatisticVariant]] = get_optional_list(data, "variants", UserStatisticVariant)
        self.rank_change_since_30_days: Optional[int] = data.get("rank_change_since_30_days")

    @staticmethod
    def calculate_recommended_difficulty(pp):
        return None if pp is None else math.pow(pp, 0.4) * 0.195

    def __repr__(self):
        return prettify(self, "pp", "global_rank", "user")


class UserStatisticVariant:
    """
    A variant ranking system.

    **Attributes**

    country_rank: Optional[int]

    global_rank: Optional[int]

    mode: :class:`GameModeStr`

    pp: int

    variant: str
        4k or 7k
    """

    __slots__ = ("country_rank", "global_rank", "mode", "pp", "variant")

    def __init__(self, data):
        self.country_rank: Optional[int] = get_required(data, "country_rank")
        self.global_rank: Optional[int] = get_required(data, "global_rank")
        self.mode: GameModeStr = GameModeStr(get_required(data, "mode"))
        self.pp: int = get_required(data, "pp")
        self.variant: str = get_required(data, "variant")


class UserStatisticsRulesets:
    """
    Object that contains statistics for each gamemode.

    **Attributes**

    osu: Optional[:class:`UserStatistics`]
        statistics for osu!standard.

    taiko: Optional[:class:`UserStatistics`]
        statistics for osu!taiko.

    fruits: Optional[:class:`UserStatistics`]
        statistics for osu!catch.

    mania: Optional[:class:`UserStatistics`]
        statistics for osu!mania.
    """

    __slots__ = ("osu", "taiko", "fruits", "mania")

    def __init__(self, data):
        self.osu: Optional[UserStatistics] = get_optional(data, "osu", UserStatistics)
        self.taiko: Optional[UserStatistics] = get_optional(data, "taiko", UserStatistics)
        self.fruits: Optional[UserStatistics] = get_optional(data, "fruits", UserStatistics)
        self.mania: Optional[UserStatistics] = get_optional(data, "mania", UserStatistics)

    def __repr__(self):
        fields = [slot for slot in self.__slots__ if getattr(self, slot, None) is not None]
        return prettify(self, *fields)


class RankHighest:
    """
    Highest rank a player achieved at any point in time.

    **Attributes**

    rank: int

    updated_at: :py:class:`datetime.datetime`
    """

    __slots__ = ("rank", "updated_at")

    def __init__(self, data):
        self.rank: int = get_required(data, "rank")
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))

    def __repr__(self):
        return prettify(self, "rank", "updated_at")


class UserAchievement:
    """
    An achievement that a user received

    **Attributes**

    achieved_at: :py:class:`datetime.datetime`

    achievement_id: int
    """

    __slots__ = ("achieved_at", "achievement_id")

    def __init__(self, data):
        self.achieved_at: datetime = fromisoformat(get_required(data, "achieved_at"))
        self.achievement_id: int = get_required(data, "achievement_id")

    def __repr__(self):
        return prettify(self, "achievement_id", "achieved_at")


class UserReplaysWatchedCount:
    """
    The count of replays watched for a month

    **Attributes**

    start_date: :py:class:`datetime.datetime`

    count: int
    """

    __slots__ = ("start_date", "count")

    def __init__(self, data):
        self.start_date: datetime = fromisoformat(get_required(data, "start_date"))
        self.count: int = get_required(data, "count")

    def __repr__(self):
        return prettify(self, "count", "start_date")


class RankHistory:
    """
    Rank history data for a user

    **Attributes**

    mode: :class:`GameModeStr`

    data: List[int]
        List of ranks from oldest to newest
    """

    __slots__ = ("mode", "data")

    def __init__(self, data):
        self.mode: GameModeStr = GameModeStr(get_required(data, "mode"))
        self.data: List[int] = get_required(data, "data")

    def __repr__(self):
        return prettify(self, "mode", "data")


class UserCover:
    """
    Cover of a user's profile

    **Attributes**

    custom_url: Optional[str]

    url: Optional[str]

    id: Optional[int]
    """

    __slots__ = ("custom_url", "url", "id")

    def __init__(self, data):
        self.custom_url: Optional[str] = get_required(data, "custom_url")
        self.url: Optional[str] = get_required(data, "url")
        self.id: Optional[int] = data.get("id")

    def __repr__(self):
        return prettify(self, "custom_url")


class Country:
    """
    Country data

    **Attributes**

    code: str

    name: str

    display: Optional[int]
    """

    __slots__ = ("code", "name", "display")

    def __init__(self, data):
        self.code: str = get_required(data, "code")
        self.name: str = get_required(data, "name")
        self.display: Optional[int] = data.get("display")

    def __repr__(self):
        return prettify(self, "code", "name")


class UserKudosu:
    """
    User kudosu data

    **Attributes**

    total: int

    available: int
    """

    __slots__ = ("total", "available")

    def __init__(self, data):
        self.total: int = get_required(data, "total")
        self.available: int = get_required(data, "available")

    def __repr__(self):
        return prettify(self, "total", "available")


class DailyChallengeUserStats:
    """
    Daily challenge stats of a user

    **Attributes**

    daily_streak_best: int

    daily_streak_current: int

    last_update: Optional[:py:class:`datetime.datetime`]

    last_weekly_streak: Optional[:py:class:`datetime.datetime`]

    playcount: int

    top_10p_placements: int

    top_50p_placements: int

    user_id: int

    weekly_streak_best: int

    weekly_streak_current: int
    """

    __slots__ = (
        "daily_streak_best",
        "daily_streak_current",
        "last_update",
        "last_weekly_streak",
        "playcount",
        "top_10p_placements",
        "top_50p_placements",
        "user_id",
        "weekly_streak_best",
        "weekly_streak_current",
    )

    def __init__(self, data):
        self.daily_streak_best: int = get_required(data, "daily_streak_best")
        self.daily_streak_current: int = get_required(data, "daily_streak_current")
        self.last_update: Optional[datetime] = get_optional(data, "last_update", fromisoformat)
        self.last_weekly_streak: Optional[datetime] = get_optional(data, "last_weekly_streak", fromisoformat)
        self.playcount: int = get_required(data, "playcount")
        self.top_10p_placements: int = get_required(data, "top_10p_placements")
        self.top_50p_placements: int = get_required(data, "top_50p_placements")
        self.user_id: int = get_required(data, "user_id")
        self.weekly_streak_best: int = get_required(data, "weekly_streak_best")
        self.weekly_streak_current: int = get_required(data, "weekly_streak_current")

    def __repr__(self):
        return prettify(self, "daily_streak_best", "daily_streak_current")


class UserTeam:
    """
    **Attributes**

    id: int

    flag_url: Optional[str]

    name: str

    short_name: str
    """

    __slots__ = ("id", "flag_url", "name", "short_name")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.flag_url: Optional[str] = get_required(data, "flag_url")
        self.name: str = get_required(data, "name")
        self.short_name: str = get_required(data, "short_name")

    def __repr__(self):
        return prettify(self, "id", "name")


class UserTeamStatistics:
    """
    **Attributes**

    team_id: int

    ruleset: :class:`GameModeInt`

    play_count: int

    ranked_score: int

    performance: int
    """

    __slots__ = ("team_id", "ruleset", "play_count", "ranked_score", "performance")

    def __init__(self, data):
        self.team_id: int = get_required(data, "team_id")
        self.ruleset: GameModeInt = GameModeInt(get_required(data, "ruleset_id"))
        self.play_count: int = get_required(data, "play_count")
        self.ranked_score: int = get_required(data, "ranked_score")
        self.performance: int = get_required(data, "performance")

    def __repr__(self):
        return prettify(self, "team_id", "performance")



from typing import Optional


__all__ = ("Path",)


class Path:
    """
    An object that defines an endpoint and various attributes of it.
    Used by :class:`Client` and passed to :class:`osu.http.HTTPHandler`
    or the async equivalent classes.

    Attributes and init parameters
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    **Attributes**

    method: str
        The http method to use (e.g. get, post, put, delete).

    path: str
        The path to the endpoint (e.g. beatmaps/727/scores)

    scope: Optional[str]
        The scope required to make this request (e.g. public, forum.write).
        None if a scope is not required.

    requires_user: bool
        True if the endpoint requires a form of user authentication. Defaults to false.

    content_type: str
        Content-Type header to use in the request. Defaults to ``application/json``.

    accept: str
        Accept header to use in the request. Defaults to ``application/json``.
    """

    __slots__ = ("method", "path", "scope", "requires_user", "content_type", "accept")

    def __init__(
        self,
        method: str,
        path: str,
        scope: Optional[str],
        requires_user=False,
        content_type="application/json",
        accept="application/json",
    ):
        self.method = method
        self.path = path
        self.scope = scope
        self.requires_user = requires_user
        self.content_type = content_type
        self.accept = accept

    @property
    def requires_auth(self):
        return self.scope is not None

    @classmethod
    def beatmap_lookup(cls):
        return cls("get", "beatmaps/lookup", "public")

    @classmethod
    def user_beatmap_score(cls, beatmap, user):
        return cls("get", f"beatmaps/{beatmap}/scores/users/{user}", "public")

    @classmethod
    def user_beatmap_scores(cls, beatmap, user):
        return cls("get", f"beatmaps/{beatmap}/scores/users/{user}/all", "public")

    @classmethod
    def beatmap_scores(cls, beatmap):
        return cls("get", f"beatmaps/{beatmap}/scores", "public")

    @classmethod
    def lazer_beatmap_scores(cls, beatmap):
        return cls("get", f"beatmaps/{beatmap}/solo-scores", "public")

    @classmethod
    def beatmap(cls, beatmap):
        return cls("get", f"beatmaps/{beatmap}", "public")

    @classmethod
    def beatmaps(cls):
        return cls("get", "beatmaps", "public")

    @classmethod
    def get_beatmap_attributes(cls, beatmap):
        return cls("post", f"beatmaps/{beatmap}/attributes", "public")

    @classmethod
    def get_beatmapset(cls, beatmapset):
        return cls("get", f"beatmapsets/{beatmapset}", "public")

    @classmethod
    def beatmapset_discussion_posts(cls):
        return cls("get", "beatmapsets/discussions/posts", "public")

    @classmethod
    def beatmapset_discussion_votes(cls):
        return cls("get", "beatmapsets/discussions/votes", "public")

    @classmethod
    def beatmapset_discussions(cls):
        return cls("get", "beatmapsets/discussions", "public")

    @classmethod
    def get_changelog_build(cls, stream, build):
        return cls("get", f"changelog/{stream}/{build}", None)

    @classmethod
    def get_changelog_listing(cls):
        return cls("get", "changelog", None)

    @classmethod
    def lookup_changelog_build(cls, changelog):
        return cls("get", f"changelog/{changelog}", None)

    @classmethod
    def get_comments(cls):
        return cls("get", "comments", None)

    @classmethod
    def get_comment(cls, comment):
        return cls("get", f"comments/{comment}", None)

    @classmethod
    def reply_topic(cls, topic):
        return cls("post", f"forums/topics/{topic}/reply", "forum.write", True)

    @classmethod
    def create_topic(cls):
        return cls("post", "forums/topics", "forum.write", True)

    @classmethod
    def get_topic_and_posts(cls, topic):
        return cls("get", f"forums/topics/{topic}", "public")

    @classmethod
    def edit_topic(cls, topic):
        return cls("patch", f"forums/topics/{topic}", "forum.write")

    @classmethod
    def edit_post(cls, post):
        return cls("patch", f"forums/posts/{post}", "forum.write")

    @classmethod
    def search(cls):
        return cls("get", "search", "public")

    @classmethod
    def get_scores(cls, room, playlist):
        return cls("get", f"rooms/{room}/playlist/{playlist}/scores", "public", True)

    @classmethod
    def get_news_listing(cls):
        return cls("get", "news", None)

    @classmethod
    def get_news_post(cls, news):
        return cls("get", f"news/{news}", None)

    @classmethod
    def revoke_current_token(cls):
        return cls("delete", "oauth/tokens/current", "public")

    @classmethod
    def get_ranking(cls, mode, type):
        return cls("get", f"rankings/{mode}/{type}", "public")

    @classmethod
    def get_spotlights(cls):
        return cls("get", "spotlights", "public")

    @classmethod
    def get_own_data(cls, mode=""):
        return cls("get", f"me/{mode}", "identify")

    @classmethod
    def get_user_kudosu(cls, user):
        return cls("get", f"users/{user}/kudosu", "public")

    @classmethod
    def get_user_scores(cls, user, type):
        return cls("get", f"users/{user}/scores/{type}", "public")

    @classmethod
    def get_user_beatmaps(cls, user, type):
        return cls("get", f"users/{user}/beatmapsets/{type}", "public")

    @classmethod
    def get_user_recent_activity(cls, user):
        return cls("get", f"users/{user}/recent_activity", "public")

    @classmethod
    def get_user(cls, user, mode=""):
        return cls("get", f"users/{user}/{mode}", "public")

    @classmethod
    def get_users(cls):
        return cls("get", "users", "public")

    @classmethod
    def lookup_users(cls):
        return cls("get", "users/lookup", "public")

    @classmethod
    def get_wiki_page(cls, locale, path):
        return cls("get", f"wiki/{locale}/{path}", None)

    @classmethod
    def get_score_by_id(cls, mode, score):
        return cls("get", f"scores/{mode}/{score}", "public")

    @classmethod
    def get_score_by_id_only(cls, score):
        return cls("get", f"scores/{score}", "public")

    @classmethod
    def get_beatmapset_events(cls):
        return cls("get", "beatmapsets/events", "public")

    @classmethod
    def get_matches(cls):
        return cls("get", "matches", "public")

    @classmethod
    def get_match(cls, match):
        return cls("get", f"matches/{match}", "public")

    @classmethod
    def get_rooms(cls, mode=""):
        return cls("get", f"rooms/{mode}", "public", True)

    @classmethod
    def get_room(cls, room):
        return cls("get", f"rooms/{room}", "public")

    @classmethod
    def get_room_leaderboard(cls, room):
        return cls("get", f"rooms/{room}/leaderboard", "public", True)

    @classmethod
    def get_seasonal_backgrounds(cls):
        return cls("get", "seasonal-backgrounds", None)

    @classmethod
    def get_replay_data(cls, mode, score):
        return cls("get", f"scores/{mode}/{score}/download", "public", False)

    @classmethod
    def get_replay_data_by_id_only(cls, score):
        return cls("get", f"scores/{score}/download", "public", False)

    @classmethod
    def get_friends(cls):
        return cls("get", "friends", "friends.read", True)

    @classmethod
    def beatmapset_search(cls):
        return cls("get", "beatmapsets/search", "public")

    @classmethod
    def chat_keepalive(cls):
        return cls("post", "chat/ack", "chat.read", True)

    @classmethod
    def create_new_pm(cls):
        return cls("post", "chat/new", "chat.write", True)

    @classmethod
    def get_channel_messages(cls, channel_id):
        return cls("get", f"chat/channels/{channel_id}/messages", "chat.read", True)

    @classmethod
    def send_message_to_channel(cls, channel_id):
        return cls("post", f"chat/channels/{channel_id}/messages", "chat.write", True)

    @classmethod
    def join_channel(cls, channel_id, user_id):
        return cls("put", f"chat/channels/{channel_id}/users/{user_id}", "chat.write_manage", True)

    @classmethod
    def leave_channel(cls, channel_id, user_id):
        return cls("delete", f"chat/channels/{channel_id}/users/{user_id}", "chat.write_manage", True)

    @classmethod
    def mark_channel_read(cls, channel_id, message_id):
        return cls("put", f"chat/channels/{channel_id}/mark-as-read/{message_id}", "chat.read", True)

    @classmethod
    def get_channel_list(cls):
        return cls("get", "chat/channels", "chat.read", True)

    @classmethod
    def create_channel(cls):
        return cls("post", "chat/channels", "chat.write_manage", True)

    @classmethod
    def get_channel(cls, channel_id):
        return cls("get", f"chat/channels/{channel_id}", "chat.read", True)

    @classmethod
    def get_all_scores(cls):
        return cls("get", "scores", "public")

    @classmethod
    def get_forums(cls):
        return cls("get", "forums", "public")

    @classmethod
    def get_forum(cls, forum_id):
        return cls("get", f"forums/{forum_id}", "public")

    @classmethod
    def get_forum_topics(cls):
        return cls("get", f"forums/topics", "public")



from time import monotonic
from typing import Callable, Optional, TYPE_CHECKING
import threading

from .scope import Scope
from .http import HTTPHandler, BaseHTTPHandler

if TYPE_CHECKING:
    from .asyncio.auth import AsynchronousAuthHandler


__all__ = ("BaseAuthHandler", "NoAuth", "FunctionalAuthHandler", "AuthUtil", "AuthHandler")


class BaseAuthHandler:
    """
    An abstract class for implementing authentication logic.
    """

    __slots__ = ("scope", "http")

    scope: Scope
    http: BaseHTTPHandler

    def get_token(self) -> Optional[str]:
        """
        Returns the access token. If the token is expired, it will be refreshed before being returned.
        """
        raise NotImplementedError()

    def has_user(self) -> bool:
        raise NotImplementedError()


class NoAuth(BaseAuthHandler):
    def __init__(self):
        self.scope = Scope()
        self.http = HTTPHandler(None)

    def get_token(self) -> Optional[str]:
        return

    def has_user(self) -> bool:
        return False


class AuthData:
    def __init__(self):
        self.refresh_token: Optional[str] = None
        self.token: Optional[str] = None
        self.expire_time: float = monotonic()

    def set_data(self, token: Optional[str], refresh_token: Optional[str], expires_in: float) -> None:
        self.token = token
        self.refresh_token = refresh_token
        self.expire_time = monotonic() + expires_in

    @property
    def has_expired(self):
        return self.expire_time - 5 <= monotonic()


class AuthUtil:
    """
    Abstract class that helps to go through the oauth process easily, as well as refresh
    tokens without the user needing to worry about it.

    Certain functions expect that the `_http` attribute is defined by classes implementing this one.

    .. note::
        If you're not authorizing a user with a url for a code, this does not apply to you.
        AuthHandler does not save refresh tokens past the program finishing.
        AuthHandler will save the refresh token to refresh the access token
        while the program is running, so make sure to save the refresh token
        before shutting down the program, so you can use it to get a valid access token
        without having the user reauthorize.

    **Init Parameters**

    client_id: int
        Client id

    client_secret: str
        Client secret

    redirect_uri: str
        Redirect uri

    scope: Optional[:class:`Scope`]
        Scopes to authorize under. Default is :func:`Scope.default`.
    """

    client_id: int
    client_secret: str
    redirect_url: str
    scope: Scope
    http: BaseHTTPHandler
    _data: AuthData
    _refresh_callback: Optional[Callable[["AuthUtil"], None]]

    SAVE_VERSION = 2

    def __init__(
        self,
        client_id: int,
        client_secret: str,
        redirect_url: Optional[str],
        scope: Optional[Scope] = None,
    ):
        if scope is None:
            scope = Scope.default()

        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_url = redirect_url
        self.scope = scope

        self._data = AuthData()
        self._refresh_callback = None

    def _get_data(self, grant_type, code=None):
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": grant_type,
        }

        data.update(
            {
                "client_credentials": {"scope": self.scope.scopes},
                "authorization_code": {"code": code, "redirect_uri": self.redirect_url},
                "refresh_token": {"refresh_token": self._data.refresh_token},
            }[grant_type]
        )

        return data

    def _handle_response(self, data):
        self._data.set_data(data["access_token"], data.get("refresh_token"), data["expires_in"])

    def get_auth_url(self, state: Optional[str] = ""):
        """
        Returns a url that a user can authorize their account at. They'll then be returned to
        the redirect_uri with a code that can be used under get_auth_token.

        **Parameters**

        state: Optional[str]
            Will be returned to the redirect_uri along with the code.
        """
        params = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_url,
            "response_type": "code",
            "scope": self.scope.scopes.replace(" ", "%20"),
            "state": state,
        }
        if not params["state"]:
            del params["state"]
        return self.http.auth_url + "?" + "&".join([f"{key}={value}" for key, value in params.items()])

    def has_user(self):
        """
        Returns whether this auth has access to endpoints requiring a user
        """
        return self.scope.has_user

    def set_refresh_callback(self, callback: Callable[["FunctionalAuthHandler"], None]):
        """
        Set a callback to be called everytime the access token is refreshed.

        **Parameters**

        callback: :class:`Callable[['AuthHandler'], None]`
        """
        self._refresh_callback = callback

    def get_save_data(self):
        """
        Get save data in json format. Can be used to easily initiate a new :class:`AuthHandler` object
        in a new session.

        **Returns**

        {

        'save_version': int,

        'client_id': int,

        'client_secret': str,

        'redirect_url': str,

        'scope': str,

        'refresh_token': str,

        'domain': str,

        }
        """
        return {
            "save_version": self.SAVE_VERSION,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_url": self.redirect_url,
            "scope": self.scope.scopes,
            "refresh_token": self._data.refresh_token,
            "domain": self.http.domain,
        }

    @classmethod
    def from_save_data(cls, save_data: dict) -> "AuthUtil":
        """
        Create a new :class:`FunctionalAuthHandler` object from save data.
        """
        save_version = save_data["save_version"]
        if save_version != cls.SAVE_VERSION:
            raise ValueError(
                f"The version of this save data ({save_version}) is not compatible "
                f"with the save data version of this AuthHandler object ({cls.SAVE_VERSION})."
            )

        return cls._from_save_data(save_data)

    @classmethod
    def _from_save_data(cls, save_data: dict) -> "AuthUtil":
        client_id = save_data["client_id"]
        client_secret = save_data["client_secret"]
        redirect_url = save_data["redirect_url"]
        scope = Scope(*save_data["scope"].split())
        auth = cls(client_id, client_secret, redirect_url, scope)
        auth.http.set_domain(save_data["domain"])
        auth._data.refresh_token = save_data["refresh_token"]
        return auth

    def set_domain(self, domain: str):
        self.http.set_domain(domain)


# backwards compatibility
FunctionalAuthHandler = AuthUtil


class AuthHandler(BaseAuthHandler, AuthUtil):
    __slots__ = ("http", "_lock")

    def __init__(
        self,
        client_id: int,
        client_secret: str,
        redirect_url: Optional[str],
        scope: Optional[Scope] = None,
    ):
        AuthUtil.__init__(self, client_id, client_secret, redirect_url, scope)

        self.http: HTTPHandler = HTTPHandler(self)
        self._lock: threading.Lock = threading.Lock()

    def has_user(self) -> bool:
        return AuthUtil.has_user(self)

    @staticmethod
    def _raise_for_status(resp):
        try:
            resp.raise_for_status()
        except Exception as exc:
            try:
                msg = resp.json()["error"]
            except:
                msg = None
            raise type(exc)(str(exc) + ": " + msg) if msg is not None else exc from None

    def get_auth_token(self, code: Optional[str] = None) -> None:
        """
        `code` parameter is not required, but without a code the scopes are restricted to
        public and delegate (more on delegation below). You can obtain a code by having
        a user authorize themselves under a url which you can get with get_auth_url.
        Read more about it under that function.

        **Client Credentials Delegation**

        Client Credentials Grant tokens may be allowed to act on behalf of the owner of the OAuth client
        (delegation) by requesting the delegate scope, in addition to other scopes supporting delegation.
        When using delegation, scopes that support delegation cannot be used together with scopes that do
        not support delegation. Delegation is only available to Chat Bots. Currently, chat.write is the only
        other scope that supports delegation.

        **Parameters**

        code: Optional[str]
            code from user authorizing at a specific url
        """
        data = self._get_data("client_credentials" if code is None else "authorization_code", code)

        response = self.http.get_auth_token(data)
        self._raise_for_status(response)
        response = response.json()

        self._handle_response(response)

    def refresh_access_token(self, refresh_token: Optional[str] = None) -> None:
        """
        This function is usually executed by HTTPHandler, but if you have a
        refresh token saved from the last session, then you can fill in the
        `refresh_token` argument which this function will use to get a valid token.

        **Parameters**

        refresh_token: Optional[str]
            A refresh token used to get a new access token.
        """
        if refresh_token:
            self._data.refresh_token = refresh_token

        data = self._get_data("client_credentials" if self._data.refresh_token is None else "refresh_token")

        response = self.http.get_auth_token(data)
        self._raise_for_status(response)
        response = response.json()

        self._handle_response(response)

        if self._refresh_callback:
            self._refresh_callback(self)

    def get_token(self) -> Optional[str]:
        with self._lock:
            if self._data.has_expired:
                self.refresh_access_token()
            return self._data.token

    @classmethod
    def from_async(cls, auth: "AsynchronousAuthHandler"):
        new_auth = cls(auth.client_id, auth.client_secret, auth.redirect_url, auth.scope)
        new_auth._data = auth._data
        new_auth._refresh_callback = auth._refresh_callback
        new_auth.http = auth.http.as_sync(new_auth)
        return new_auth

    def as_async(self) -> "AsynchronousAuthHandler":
        """
        Returns an asynchronous auth handler.
        Credentials are shared and updates in one apply to the other.
        This method primarily exists for running the library's tests.

        NOTE: Using both auth handlers in different threads at the same time is not safe.
        """
        from .asyncio.auth import AsynchronousAuthHandler

        return AsynchronousAuthHandler.from_sync(self)



__all__ = ("Scope",)


class Scope:
    """
    Scope object for telling the program what scopes you are using

    **Valid scopes**

    public
        Allows reading of publicly available data on behalf of the user.
    identify (default)
        Allows reading of the public profile of the user (/me).
    friends.read
        Allows reading of the user's friend list.
    forum.write
        Allows creating and editing forum posts on a user's behalf.
    delegate
        Allows acting as the owner of a client; only available for Client Credentials Grant.
    chat.write
        Allows sending chat messages on a user's behalf.
    """

    __slots__ = ("scopes", "scopes_list")
    valid_scopes = [
        "chat.read",
        "chat.write",
        "chat.write_manage",
        "delegate",
        "forum.write",
        "friends.read",
        "identify",
        "public",
    ]

    def __init__(self, *scopes):
        for scope in scopes:
            if scope not in self.valid_scopes and scope != "*":
                raise NameError(
                    f"{scope} is not a valid scope. The valid scopes consist of " f"{','.join(self.valid_scopes)}"
                )
        self.scopes = " ".join(scopes)
        self.scopes_list = list(scopes)

    @classmethod
    def default(cls):
        """Initializes with ``public`` scope"""
        return cls("public")

    @classmethod
    def identify(cls):
        """Initializes with ``public`` and ``identify`` scopes"""
        return cls("public", "identify")

    def __iter__(self):
        return iter(self.scopes_list)

    def __len__(self):
        return len(self.scopes_list)

    def __str__(self):
        return str(self.scopes_list)

    def __contains__(self, scope: str):
        return scope in self.scopes_list or "*" in self.scopes_list

    @property
    def has_user(self):
        return any((scope.lower() != "public" for scope in self.scopes_list))



import setuptools
import re


with open("requirements.txt") as f:
    requirements = f.read().splitlines()

readme = ''
with open('README.rst') as f:
    readme = f.read()

version = ''
with open('osu/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)
    
readme = ''
with open('README.rst') as f:
    readme = f.read()

project_urls = {
    "Bug Tracker": "https://github.com/Sheepposu/osu.py/issues",
    "osu.py documentation": "https://osupy.readthedocs.io/",
    "osu!api v2 documentation": "https://osu.ppy.sh/docs/index.html",
    "Code examples": "https://github.com/Sheepposu/osu.py/tree/main/examples"
}
classifiers = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Natural Language :: English",
    "Intended Audience :: Developers"
]

extra_require = {
    "async": ["aiohttp>=3.9.2,<4"],
    "replay": ["osrparse>=7.0.1,<8"],
    "notifications": ["websockets>=13.1,<14"],
    "tests": [
        "pytest>=8.3.3,<9",
        "pytest-asyncio>=0.24.0,<1",
        "pytest-dependency>=0.6.0,<1"
    ]
}

packages = [
    'osu',
    'osu.asyncio',
    'osu.objects',
]

setuptools.setup(
    name="osu.py",
    version=version,
    packages=packages,
    author="Sheepposu",
    description="API Wrapper for osu!api v2 written in Python.",
    long_description=readme,
    long_description_content_type="text/x-rst",
    install_requires=requirements,
    extras_require=extra_require,
    project_urls=project_urls,
    classifiers=classifiers,
    python_requires=">=3.8.0",
    license="MIT",
    url="https://github.com/Sheepposu/osu.py",
)



from typing import Dict, Optional, List, TYPE_CHECKING, Union

from ..enums import (
    GameModeInt,
    PlaylistQueueMode,
    RealTimeQueueMode,
    RoomCategory,
    RoomType,
    RoomStatus,
)
from ..util import prettify, get_optional, get_optional_list, get_required, fromisoformat
from .user import UserCompact
from .beatmap import BeatmapCompact
from .score import ScoreDataStatistics, LazerMod


if TYPE_CHECKING:
    from datetime import datetime


__all__ = (
    "MultiplayerScore",
    "MultiplayerScores",
    "MultiplayerScoresAround",
    "Room",
    "UserScoreAggregate",
    "PlaylistItemStats",
    "PlaylistItem",
)


class MultiplayerScore:
    """
    Score data.

    **Attributes**

    id: int

    user_id: int

    room_id: int

    playlist_item_id: int

    beatmap_id: int

    rank: str

    total_score: int

    accuracy: float

    max_combo: int

    mods: List[:class:`LazerMod`]

    statistics: :class:`ScoreDataStatistics`

    passed: bool

    position: Optional[int]

    scores_around: Optional[:class:`MultiplayerScoresAround`]
        Scores around the specified score.

    user: :class:`User`
    """

    __slots__ = (
        "id",
        "user_id",
        "room_id",
        "playlist_item_id",
        "beatmap_id",
        "rank",
        "total_score",
        "accuracy",
        "max_combo",
        "mods",
        "statistics",
        "passed",
        "position",
        "scores_around",
        "user",
    )

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.user_id: int = get_required(data, "user_id")
        self.room_id: int = get_required(data, "room_id")
        self.playlist_item_id: int = get_required(data, "playlist_item_id")
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.rank: str = get_required(data, "rank")
        self.total_score: int = get_required(data, "total_score")
        self.accuracy: float = get_required(data, "accuracy")
        self.max_combo: int = get_required(data, "max_combo")
        self.mods: Optional[List[LazerMod]] = get_optional_list(data, "mods", LazerMod)
        self.statistics: ScoreDataStatistics = ScoreDataStatistics(get_required(data, "statistics"))
        self.passed: bool = get_required(data, "passed")
        self.position: Optional[int] = data.get("position")
        self.scores_around: Optional[MultiplayerScoresAround] = get_optional(
            data, "scores_around", MultiplayerScoresAround
        )
        self.user: Optional[UserCompact] = get_optional(data, "user", UserCompact)

    def __repr__(self):
        return prettify(self, "user", "position")


class MultiplayerScores:
    """
    An object which contains scores and related data for fetching next page of the result.

    **Attributes**

    cursor: str
        To be used to fetch the next page.

    params: :class:`dict`
        To be used to fetch the next page.

    scores: List[:class:`MultiplayerScore`]

    total: Optional[int]
        Index only. Total scores of the specified playlist item.

    user_score: Optional[:class:`MultiplayerScore`]
        Index only. Score of the accessing user if exists.
    """

    __slots__ = ("cursor", "params", "scores", "total", "user_score")

    def __init__(self, data):
        self.cursor: str = get_required(data, "cursor")
        self.params: Dict = get_required(data, "params")
        self.scores: List[MultiplayerScore] = list(map(MultiplayerScore, get_required(data, "scores")))
        self.total: Optional[int] = data.get("total")
        self.user_score: Optional[MultiplayerScore] = get_optional(data, "user_score", MultiplayerScore)

    def __repr__(self):
        return prettify(self, "total", "scores")


class MultiplayerScoresAround:
    """
    **Attributes**

    higher: :class:`MultiplayerScores`

    lower: :class:`MultiplayerScores`
    """

    __slots__ = ("higher", "lower")

    def __init__(self, data):
        self.higher: MultiplayerScores = MultiplayerScores(get_required(data, "higher"))
        self.lower: MultiplayerScores = MultiplayerScores(get_required(data, "lower"))

    def __repr__(self):
        return prettify(self, "higher", "lower")


class Room:
    """
    :func:`osu.Client.get_rooms` and :func:`osu.Client.get_room` endpoints include host, playlist,
    and recent_participants attributes. In addition, the :class:`PlaylistItem` objects in playlist
    include the beatmap attribute and the beatmap attribute has the beatmapset, checksum, and
    max_combo attributes.

    **Attributes**

    id: int

    name: str

    category: :class:`RoomCategory`

    status: :class:`RoomStatus`

    type: :class:`RoomType`

    realtime_type: :class:`RealTimeType`
        Only applicable when type is RoomType.REALTIME

    user_id: int

    starts_at: Optional[:py:class:`datetime.datetime`]

    ends_at: Optional[:py:class:`datetime.datetime`]

    max_attempts: int

    participant_count: int

    channel_id: int

    active: bool

    has_password: bool

    queue_mode: Union[:class:`RealTimeQueueMode`, :class:`PlaylistQueueMode`]
        Type depends on the room type. :class:`PlaylistQueueMode` for type :class:`RoomType`.PLAYLIST
        and :class:`RealTimeQueueMode` otherwise.

    current_playlist_item: Optional[:class:`PlaylistItem`]

    current_user_score: Optional[:class:`UserScoreAggregate`]

    difficulty_range: Optional[Dict[str, int]]
        When not none, is a dictionary containing keys "max" and "min"

    host: Optional[:class:`UserCompact`]

    playlist: Optional[List[:class:`PlaylistItem`]]

    playlist_item_stats: Optional[:class:`PlaylistItemStats`]

    recent_participants: Optional[List[:class:`UserCompact`]]

    scores: Optional[:class:`MultiplayerScore`]
    """

    __slots__ = (
        "id",
        "name",
        "category",
        "status",
        "type",
        "user_id",
        "starts_at",
        "ends_at",
        "max_attempts",
        "participant_count",
        "channel_id",
        "active",
        "has_password",
        "queue_mode",
        "current_playlist_item",
        "current_user_score",
        "difficulty_range",
        "host",
        "playlist",
        "playlist_item_stats",
        "recent_participants",
        "scores",
    )

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.name: str = get_required(data, "name")
        self.category: RoomCategory = RoomCategory(get_required(data, "category"))
        self.status: RoomStatus = RoomStatus(get_required(data, "status"))
        self.type: RoomType = RoomType(get_required(data, "type"))
        self.user_id: int = get_required(data, "user_id")
        self.starts_at: datetime = fromisoformat(get_required(data, "starts_at"))
        self.ends_at: Optional[datetime] = get_optional(data, "ends_at", fromisoformat)
        self.max_attempts: Optional[int] = get_required(data, "max_attempts")
        self.participant_count: int = get_required(data, "participant_count")
        self.channel_id: Optional[int] = get_required(data, "channel_id")
        self.active: bool = get_required(data, "active")
        self.has_password: bool = get_required(data, "has_password")
        self.queue_mode: Union[RealTimeQueueMode, PlaylistQueueMode] = (
            RealTimeQueueMode if self.type != RoomType.PLAYLISTS else PlaylistQueueMode
        )(get_required(data, "queue_mode"))

        self.current_playlist_item: Optional[PlaylistItem] = get_optional(data, "current_playlist_item", PlaylistItem)
        self.current_user_score: Optional[UserScoreAggregate] = get_optional(
            data, "current_user_score", UserScoreAggregate
        )
        self.difficulty_range: Optional[Dict[str, int]] = data.get("difficulty_range")  # {min: int, max: int}
        self.host: Optional[UserCompact] = get_optional(data, "host", UserCompact)
        self.playlist: Optional[List[PlaylistItem]] = get_optional_list(data, "playlist", PlaylistItem)
        self.playlist_item_stats: Optional[PlaylistItemStats] = get_optional(
            data, "playlist_item_stats", PlaylistItemStats
        )
        self.recent_participants: Optional[List[UserCompact]] = get_optional_list(
            data, "recent_participants", UserCompact
        )
        self.scores: Optional[List[MultiplayerScore]] = get_optional_list(data, "scores", MultiplayerScore)

    def __repr__(self):
        return prettify(self, "name", "type")


class UserScoreAggregate:
    """
    **Attributes**

    accuracy: float

    attempts: int

    completed: int

    pp: float

    room_id: int

    total_score: int

    user_id: int

    playlist_item_attempts: Optional[Dict[str, int]]
        contains keys "attempts" and "id".

    position: Optional[int]
        user rank

    user: Optional[:class:`UserCompact`]
    """

    __slots__ = (
        "accuracy",
        "attempts",
        "completed",
        "pp",
        "room_id",
        "total_score",
        "user_id",
        "playlist_item_attempts",
        "position",
        "user",
    )

    def __init__(self, data):
        self.accuracy: float = get_required(data, "accuracy")
        self.attempts: int = get_required(data, "attempts")
        self.completed: int = get_required(data, "completed")
        self.pp: float = get_required(data, "pp")
        self.room_id: int = get_required(data, "room_id")
        self.total_score: int = get_required(data, "total_score")
        self.user_id: int = get_required(data, "user_id")
        self.playlist_item_attempts: Optional[Dict[str, int]] = data.get(
            "playlist_item_attempts"
        )  # {attempts: int, id: int}
        self.position: Optional[int] = data.get("position")
        self.user: Optional[UserCompact] = get_optional(data, "user", UserCompact)

    def __repr__(self):
        attributes = ("total_score", "user" if self.user is not None else "user_id")
        return prettify(self, *attributes)


class PlaylistItemStats:
    """
    **Attributes**

    count_active: int

    count_total: int

    ruleset_ids: List[:class:`GameModeInt`]
    """

    __slots__ = ("count_active", "count_total", "ruleset_ids")

    def __init__(self, data):
        self.count_active: int = get_required(data, "count_active")
        self.count_total: int = get_required(data, "count_total")
        self.ruleset_ids: List[GameModeInt] = list(map(GameModeInt, get_required(data, "ruleset_ids")))

    def __repr__(self):
        return prettify(self, "count_active", "count_total")


class PlaylistItem:
    """
    **Attributes**

    id: int

    room_id: int

    beatmap_id: int

    ruleset_id: :class:`GameModeInt`

    allowed_mods: List[:class:`LazerMod`]

    required_mods: List[:class:`LazerMod`]

    expired: bool

    owner_id: int

    playlist_order: Optional[int]

    played_at: Optional[:py:class:`datetime.datetime`]

    beatmap: Optional[:class:`BeatmapCompact`]
    """

    __slots__ = (
        "id",
        "room_id",
        "beatmap_id",
        "ruleset_id",
        "allowed_mods",
        "required_mods",
        "expired",
        "owner_id",
        "playlist_order",
        "played_at",
        "beatmap",
    )

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.room_id: int = get_required(data, "room_id")
        self.beatmap_id: int = get_required(data, "beatmap_id")
        self.ruleset_id: int = GameModeInt(get_required(data, "ruleset_id"))
        self.allowed_mods: List[LazerMod] = list(map(LazerMod, get_required(data, "allowed_mods")))
        self.required_mods: List[LazerMod] = list(map(LazerMod, get_required(data, "required_mods")))
        self.expired: bool = get_required(data, "expired")
        self.owner_id: int = get_required(data, "owner_id")
        self.playlist_order: Optional[int] = get_required(data, "playlist_order")
        self.played_at: Optional[datetime] = get_optional(data, "played_at", fromisoformat)
        self.beatmap: Optional[BeatmapCompact] = get_optional(data, "beatmap", BeatmapCompact)

    def __repr__(self):
        attributes = ("id", "beatmap" if self.beatmap is not None else "beatmap_id")
        return prettify(self, *attributes)



from typing import Optional, List, TYPE_CHECKING

from ..util import prettify, get_optional, get_required, get_optional_list, fromisoformat
from ..enums import ForumTopicType


if TYPE_CHECKING:
    from datetime import datetime


__all__ = ("Forum", "ForumPost", "TextFormat", "ForumTopic", "Poll", "PollOption")


class Forum:
    """
    **Attributes**

    id: int

    name: str

    description: str

    subforums: Optional[List[:class:`Forum`]]
    """

    __slots__ = ("id", "name", "description", "subforums")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.name: str = get_required(data, "name")
        self.description: str = get_required(data, "description")
        self.subforums: List[Forum] = get_optional_list(data, "subforums", Forum)

    def __repr__(self):
        return prettify(self, "id", "name", "subforums")


class ForumPost:
    """
    **Attributes**

    created_at: :py:class:`datetime.datetime`

    deleted_at: Optional[:py:class:`datetime.datetime`]

    edited_at: Optional[:py:class:`datetime.datetime`]

    edited_by_id: Optional[int]

    forum_id: int

    id: int

    topic_id: int

    user_id: int

    body: Optional[:class:`TextFormat`]
    """

    __slots__ = (
        "created_at",
        "deleted_at",
        "edited_at",
        "edited_by_id",
        "forum_id",
        "id",
        "topic_id",
        "user_id",
        "body",
    )

    def __init__(self, data):
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", fromisoformat)
        self.edited_at: Optional[datetime] = get_optional(data, "edited_at", fromisoformat)
        self.edited_by_id: Optional[int] = data.get("edited_by_id")
        self.forum_id: int = get_required(data, "forum_id")
        self.id: int = get_required(data, "id")
        self.topic_id: int = get_required(data, "topic_id")
        self.user_id: int = get_required(data, "user_id")
        self.body: Optional[TextFormat] = get_optional(data, "body", TextFormat)

    def __repr__(self):
        return prettify(self, "user_id", "topic_id")


class TextFormat:
    """
    A page that has html and raw data

    **Attributes**

    html: str

    raw: str
    """

    __slots__ = ("html", "raw")

    def __init__(self, data):
        self.html: str = get_required(data, "html")
        for attr in ("raw", "bbcode", "markdown"):
            if attr in data:
                self.raw: str = data[attr]
                break
        if self.raw is None:
            raise KeyError("data did not contain a raw value")

    def __repr__(self):
        return prettify(self, "raw")


class ForumTopic:
    """
    **Attributes**

    created_at: :py:class:`datetime.datetime`

    deleted_at: Optional[:py:class:`datetime.datetime`]

    first_post_id: int

    forum_id: int

    id: int

    is_locked: bool

    last_post_id: int

    poll: Optional[:class:`Poll`]

    post_count: int

    title: str

    type: :class:`ForumTopicType`

    updated_at: :py:class:`datetime.datetime`

    user_id: int
    """

    __slots__ = (
        "created_at",
        "deleted_at",
        "first_post_id",
        "forum_id",
        "id",
        "is_locked",
        "last_post_id",
        "poll",
        "post_count",
        "title",
        "type",
        "updated_at",
        "user_id",
    )

    def __init__(self, data):
        self.created_at: datetime = fromisoformat(get_required(data, "created_at"))
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", fromisoformat)
        self.first_post_id: int = get_required(data, "first_post_id")
        self.forum_id: int = get_required(data, "forum_id")
        self.id: int = get_required(data, "id")
        self.is_locked: bool = get_required(data, "is_locked")
        self.last_post_id: int = get_required(data, "last_post_id")
        self.poll: Optional[Poll] = get_optional(data, "poll", Poll)
        self.post_count: int = get_required(data, "post_count")
        self.title: str = get_required(data, "title")
        self.type: ForumTopicType = ForumTopicType(get_required(data, "type"))
        self.updated_at: datetime = fromisoformat(get_required(data, "updated_at"))
        self.user_id: int = get_required(data, "user_id")

    def __repr__(self):
        return prettify(self, "user_id", "title")


class Poll:
    """
    **Attributes**

    allow_vote_change: bool

    ended_at: Optional[:py:class:`datetime.datetime`]

    hide_incomplete_results: bool

    last_vote_at: Optional[:py:class:`datetime.datetime`]

    max_votes: int

    options: List[:class:`PollOption`]

    started_at: :py:class:`datetime.datetime`

    title: :class:`TextFormat`

    total_vote_count: int
    """

    __slots__ = (
        "allow_vote_change",
        "ended_at",
        "hide_incomplete_results",
        "last_vote_at",
        "max_votes",
        "options",
        "started_at",
        "title",
        "total_vote_count",
    )

    def __init__(self, data):
        self.allow_vote_change: bool = get_required(data, "allow_vote_change")
        self.ended_at: Optional[datetime] = get_optional(data, "ended_at", fromisoformat)
        self.hide_incomplete_results: bool = get_required(data, "hide_incomplete_results")
        self.last_vote_at: Optional[datetime] = get_optional(data, "last_vote_at", fromisoformat)
        self.max_votes: int = get_required(data, "max_votes")
        self.options: List[PollOption] = list(map(PollOption, get_required(data, "options")))
        self.started_at: datetime = fromisoformat(get_required(data, "started_at"))
        self.title: TextFormat = TextFormat(get_required(data, "title"))
        self.total_vote_count: int = get_required(data, "total_vote_count")

    def __repr__(self):
        return prettify(self, "title", "options")


class PollOption:
    """
    **Attributes**

    id: int

    text: :class:`TextFormat`

    vote_count: Optional[int]
        Not present if the poll is incomplete and results are hidden.
    """

    __slots__ = ("id", "text", "vote_count")

    def __init__(self, data):
        self.id: int = get_required(data, "id")
        self.text: TextFormat = TextFormat(get_required(data, "text"))
        self.vote_count: Optional[int] = get_required(data, "vote_count")

    def __repr__(self):
        return prettify(self, "text", "vote_count")



from typing import Optional, Dict

from ..util import prettify, get_optional, get_required
from ..enums import ObjectType, GameModeStr


__all__ = (
    "BeatmapsetDiscussionPermissions",
    "BeatmapsetPermissions",
    "ChatChannelUserAttributes",
    "ScoreUserAttributes",
    "CurrentUserPin",
    "CommentableMetaAttributes",
)


class BeatmapsetDiscussionPermissions:
    """
    A user's permissions in a beatmapset discussion

    **Attributes**

    can_destroy: bool

    can_reopen: bool

    can_moderate_kudosu: bool

    can_resolve: bool

    vote_score: int
    """

    __slots__ = (
        "can_destroy",
        "can_reopen",
        "can_moderate_kudosu",
        "can_resolve",
        "vote_score",
    )

    def __init__(self, data):
        self.can_destroy: bool = get_required(data, "can_destroy")
        self.can_reopen: bool = get_required(data, "can_reopen")
        self.can_moderate_kudosu: bool = get_required(data, "can_moderate_kudosu")
        self.can_resolve: bool = get_required(data, "can_resolve")
        self.vote_score: int = get_required(data, "vote_score")

    def __repr__(self):
        return prettify(self, "vote_score")


class BeatmapsetPermissions:
    """
    User permissions on a beatmapset

    **Attributes**

    can_beatmap_update_owner: bool

    can_delete: bool

    can_edit_metadata: bool

    can_edit_offset: bool

    can_edit_tags: bool

    can_hype: bool

    can_hype_reason: str

    can_love: bool

    can_remove_from_loved: bool

    is_watching: bool

    new_hype_time: Optional[str]

    nomination_modes: Dict[:class:`GameModeStr`, str]
        Values are either "full" or "limited".

    remaining_hype: int
    """

    __slots__ = (
        "can_beatmap_update_owner",
        "can_delete",
        "can_edit_metadata",
        "can_edit_offset",
        "can_edit_tags",
        "can_hype",
        "can_hype_reason",
        "can_love",
        "can_remove_from_loved",
        "is_watching",
        "new_hype_time",
        "nomination_modes",
        "remaining_hype",
    )

    def __init__(self, data):
        self.can_beatmap_update_owner: bool = get_required(data, "can_beatmap_update_owner")
        self.can_delete: bool = get_required(data, "can_delete")
        self.can_edit_metadata: bool = get_required(data, "can_edit_metadata")
        self.can_edit_offset: bool = get_required(data, "can_edit_offset")
        self.can_edit_tags: bool = get_required(data, "can_edit_tags")
        self.can_hype: bool = get_required(data, "can_hype")
        self.can_hype_reason: str = get_required(data, "can_hype_reason")
        self.can_love: bool = get_required(data, "can_love")
        self.can_remove_from_loved: bool = get_required(data, "can_remove_from_loved")
        self.is_watching: bool = get_required(data, "is_watching")
        self.new_hype_time: Optional[str] = get_required(data, "new_hype_time")
        self.nomination_modes: Dict[GameModeStr, str] = get_optional(
            data,
            "nomination_modes",
            lambda value: dict(map(lambda item: (GameModeStr(item[0]), item[1]), value.items())),
        )
        self.remaining_hype: int = get_required(data, "remaining_hype")


class ChatChannelUserAttributes:
    """
    Data about a user related to a chat channel.

    **Attributes**

    can_message: bool

    can_message_error: Optional[str]

    last_read_id: Optional[int]
    """

    __slots__ = ("can_message", "can_message_error", "last_read_id")

    def __init__(self, data):
        self.can_message: bool = get_required(data, "can_message")
        self.can_message_error: Optional[str] = get_required(data, "can_message_error")
        self.last_read_id: Optional[int] = get_required(data, "last_read_id")

    def __repr__(self):
        return prettify(self, "can_message", "last_read_id")


class ScoreUserAttributes:
    """
    Gives info about a score related to the current user

    **Attributes**

    pin: Optional[:class:`CurrentUserPin`]
    """

    __slots__ = ("pin",)

    def __init__(self, data):
        self.pin: Optional[CurrentUserPin] = get_optional(data, "pin", CurrentUserPin)


class CurrentUserPin:
    """
    Gives info about a score related to if it's pinned or not

    **Attributes**

    is_pinned: bool

    score_id: int

    score_type: Optional[:class:`ObjectType`]
    """

    __slots__ = ("is_pinned", "score_id", "score_type")

    def __init__(self, data):
        self.is_pinned: bool = get_required(data, "is_pinned")
        self.score_id: int = get_required(data, "score_id")
        self.score_type: Optional[ObjectType] = get_optional(data, "score_type", ObjectType)


class CommentableMetaAttributes:
    """
    User attributes for a :class:`CommentableMeta` object

    **Attributes**

    can_new_comment_reason: Optional[str]
    """

    __slots__ = ("can_new_comment_reason",)

    def __init__(self, data):
        self.can_new_comment_reason = get_required(data, "can_new_comment_reason")
